# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QLabel, QLineEdit, QPushButton, QHBoxLayout, QGridLayout, QDialog, QRadioButton


class newPortDlg(QDialog):
    def __init__(self, block, parent=None):
        super(newPortDlg, self).__init__(parent)
        self.block = block
        nameLabel = QLabel("Name:")
        objectLabel = QLabel("Object:" + str(self.block))
        self.le = QLineEdit("0")

        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.okButton)
        buttonLayout.addWidget(self.cancelButton)

        radioButtonLayout = QHBoxLayout()
        self.oButton = QRadioButton("Output")
        self.iButton = QRadioButton("Input")
        radioButtonLayout.addWidget(self.iButton)
        radioButtonLayout.addWidget(self.oButton)

        layout = QGridLayout()
        layout.addWidget(nameLabel, 0, 0)
        layout.addWidget(self.le, 0, 1)
        layout.addLayout(radioButtonLayout, 1, 0, 2, 0)
        layout.addLayout(buttonLayout, 3, 0, 2, 0)
        layout.addWidget(objectLabel, 5, 0, 1, 2)  # Only for debug (Why do I need a 3 here instead of a 2 for int:row?)
        self.setLayout(layout)

        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)
        self.setWindowTitle("Change connection name")
        self.show()

    def acceptedEdit(self):
        if self.le.text() != "":
            height = self.le.text()

            if self.oButton.isChecked():
                io = "o"
            else:
                io = "i"

            self.block.addPort(io, height)

            self.close()

    def cancel(self):
        self.close()



# pylint: skip-file
# type: ignore

import typing as _tp

from PyQt5.QtWidgets import (
    QLabel,
    QLineEdit,
    QGridLayout,
    QHBoxLayout,
    QListWidget,
    QPushButton,
    QSpacerItem,
    QVBoxLayout,
    QRadioButton,
    QDialog,
    QTabWidget,
    QWidget,
    QMessageBox,
    QListWidgetItem,
)
import PyQt5.QtCore as _qtc

import trnsysGUI.directPortPair as _dpp
import trnsysGUI.side as _side
from trnsysGUI.modifyRelativeHeightsDialog import ModifyRelativeHeightsDialog


class ConfigureStorageDialog(QDialog):
    WIDTH_INCREMENT = 10
    HEIGHT_INCREMENT = 100

    def __init__(self, storage, parent):
        super(ConfigureStorageDialog, self).__init__(parent)
        self.parent = parent
        self.storage = storage
        self.n = 0
        self.m = 0

        spacerHeight = 15

        self.tabs = QTabWidget()
        self.tab1 = QWidget()
        self.tab2 = QWidget()

        self.tabs.addTab(self.tab1, "Heat exchangers")
        self.tabs.addTab(self.tab2, "Direct ports")

        h0 = QHBoxLayout()
        description = QLabel("Please configure the storage tank:")
        exportButton = QPushButton("Export ddck")
        exportButton.clicked.connect(self.storage.exportDck)

        h0.addWidget(description)
        h0.addWidget(exportButton)

        tankNameLabel = QLabel()
        tankNameLabel.setText("<b>Tank name: </b>")
        self.le = QLineEdit(self.storage.label.toPlainText())

        gl = QGridLayout()
        hxsLabel = QLabel("<b>Heat Exchangers </b>")
        gl.addWidget(hxsLabel, 0, 0, 1, 2)

        qhbL = QHBoxLayout()

        self.leftHeatExchangersItemListWidget = QListWidget()
        qhbL.addWidget(self.leftHeatExchangersItemListWidget)

        self.rightHeatExchangersItemListWidget = QListWidget()
        qhbL.addWidget(self.rightHeatExchangersItemListWidget)

        offsetLabel = QLabel("Height offsets in percent")
        offsetLeILabel = QLabel("Input (upper port): ")
        offsetLeOLabel = QLabel("Output (lower port):")
        self.offsetLeI = QLineEdit("0")
        self.offsetLeO = QLineEdit("0")
        self.lButton = QRadioButton("Left side")
        self.rButton = QRadioButton("Right side")

        gl.addWidget(offsetLabel, 3, 0, 1, 1)
        gl.addWidget(offsetLeILabel, 4, 0, 1, 1)
        gl.addWidget(self.offsetLeI, 5, 0, 1, 3)
        gl.addWidget(offsetLeOLabel, 6, 0, 1, 1)
        gl.addWidget(self.offsetLeO, 7, 0, 1, 3)

        self.hxNameLe = QLineEdit()
        self.hxNameLe.setPlaceholderText("Enter a name...")
        gl.addWidget(self.hxNameLe, 8, 0, 1, 3)

        gl.addWidget(self.lButton, 9, 0, 1, 1)
        gl.addWidget(self.rButton, 9, 2, 1, 1)

        addButton = QPushButton("Add...")
        addButton.clicked.connect(self.addHx)
        gl.addWidget(addButton, 10, 0, 1, 1)
        removeButton = QPushButton("Remove...")
        removeButton.clicked.connect(self.removeHxL)
        removeButton.clicked.connect(self.removeHxR)
        gl.addWidget(removeButton, 10, 1, 1, 1)
        modifyButton = QPushButton("Modify")
        modifyButton.clicked.connect(self.modifyHx)
        gl.addWidget(modifyButton, 10, 2, 1, 1)
        spaceHx = QSpacerItem(self.width(), spacerHeight)
        gl.addItem(spaceHx, 11, 0, 1, 2)

        manPortLay = QVBoxLayout()
        qhbL2 = QHBoxLayout()

        self._leftDirectPortPairsItemListWidget = QListWidget()
        qhbL2.addWidget(self._leftDirectPortPairsItemListWidget)

        self._rightDirectPortPairsItemListWidget = QListWidget()
        qhbL2.addWidget(self._rightDirectPortPairsItemListWidget)

        manPortLay.addLayout(qhbL2)

        manPortLabel = QLabel("<b>Set port manually</b>")
        manPortLabel2 = QLabel("Enter height in percent: ")
        portlabelUpper = QLabel("Inlet")
        self.manPortLeI = QLineEdit("0")
        portlabelLower = QLabel("Outlet")
        self.manPortLeO = QLineEdit("0")

        qhbl3 = QHBoxLayout()
        self.manlButton = QRadioButton("Left side")
        self.manrButton = QRadioButton("Right side")
        qhbl3.addWidget(self.manlButton)
        qhbl3.addWidget(self.manrButton)

        self.manAddButton = QPushButton("Add (manual) ports")
        self.manAddButton.clicked.connect(self.manAddPortPair)

        self.manRemovebutton = QPushButton("Remove ports")
        self.manRemovebutton.clicked.connect(self.manRemovePortPairLeft)
        self.manRemovebutton.clicked.connect(self.manRemovePortPairRight)

        self.modifyPortButton = QPushButton("Modify")
        self.modifyPortButton.clicked.connect(self.modifyPort)

        addRemoveButtons = QHBoxLayout()
        addRemoveButtons.addWidget(self.manAddButton)
        addRemoveButtons.addWidget(self.manRemovebutton)
        addRemoveButtons.addWidget(self.modifyPortButton)

        manPortLay.addWidget(manPortLabel)
        manPortLay.addWidget(manPortLabel2)
        manPortLay.addWidget(portlabelUpper)
        manPortLay.addWidget(self.manPortLeI)
        manPortLay.addWidget(portlabelLower)
        manPortLay.addWidget(self.manPortLeO)
        manPortLay.addLayout(qhbl3)
        manPortLay.addLayout(addRemoveButtons)

        increaseSizeButton = QPushButton("Increase size")
        decreaseSizeButton = QPushButton("Decrease size")
        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        increaseSizeButton.clicked.connect(self.incrSize)
        decreaseSizeButton.clicked.connect(self.decrSize)
        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)

        L = QVBoxLayout()
        L.addLayout(h0)
        L.addWidget(tankNameLabel)
        L.addWidget(self.le)

        t1Layout = QVBoxLayout()
        t1Layout.addLayout(qhbL)
        t1Layout.addLayout(gl)

        self.tab1.setLayout(t1Layout)

        self.tab2.setLayout(manPortLay)

        L.addWidget(self.tabs)

        L.addWidget(increaseSizeButton)
        L.addWidget(decreaseSizeButton)
        L.addWidget(self.okButton)
        L.addWidget(self.cancelButton)

        self.setLayout(L)

        self._loadHeatExchangers()
        self._loadDirectPortPairs()

        # This is to ensure that only one list element is selected
        self.rightHeatExchangersItemListWidget.setSelectionMode(1)
        self.leftHeatExchangersItemListWidget.setSelectionMode(1)
        self.rightHeatExchangersItemListWidget.clicked.connect(self.listWRClicked)
        self.leftHeatExchangersItemListWidget.clicked.connect(self.listWLClicked)

        self._rightDirectPortPairsItemListWidget.setSelectionMode(1)
        self._leftDirectPortPairsItemListWidget.setSelectionMode(1)
        self._rightDirectPortPairsItemListWidget.clicked.connect(self.listWR2Clicked)
        self._leftDirectPortPairsItemListWidget.clicked.connect(self.listWL2Clicked)

        self.show()

    def _loadHeatExchangers(self):
        for h in self.storage.heatExchangers:
            listItem = self._getHeatExchangerListItemText(h)
            if h.sSide == 0:
                self.leftHeatExchangersItemListWidget.addItem(listItem)
            if h.sSide == 2:
                self.rightHeatExchangersItemListWidget.addItem(listItem)

    @staticmethod
    def _getHeatExchangerListItemText(h):
        return (
            h.displayName
            + ", y_offset = "
            + "%d" % int(h.relativeInputHeight * 100)
            + "%"
            + " to "
            + "%d" % int(h.relativeOutputHeight * 100)
            + "%"
        )

    def _loadDirectPortPairs(self):
        directPortPair: _dpp.DirectPortPair
        for directPortPair in self.storage.directPortPairs:
            itemText = self._getDirectPortPairListItemText(directPortPair)
            item = QListWidgetItem(itemText)
            item.setData(_qtc.Qt.UserRole, directPortPair)

            if directPortPair.isOnLeftSide:
                self._leftDirectPortPairsItemListWidget.addItem(item)
            else:
                self._rightDirectPortPairsItemListWidget.addItem(item)

    @staticmethod
    def _getDirectPortPairListItemText(directPortPair: _dpp.DirectPortPair):
        return (
            "Port pair from "
            + "%d%%" % int(directPortPair.relativeInputHeight * 100)
            + " to "
            + "%d%%" % int(directPortPair.relativeOutputHeight * 100)
        )

    def listWLClicked(self):
        self.rightHeatExchangersItemListWidget.clearSelection()

    def listWRClicked(self):
        self.leftHeatExchangersItemListWidget.clearSelection()

    def listWL2Clicked(self):
        self._rightDirectPortPairsItemListWidget.clearSelection()

    def listWR2Clicked(self):
        self._leftDirectPortPairsItemListWidget.clearSelection()

    def addHx(self):
        """
        Checks whether the inputs are in the correct range [0,100] and in order, and calls the function for creating a
        HeatExchanger on the respective side.

        Returns
        -------
        """
        if (
            self.minOffsetDistance()
            and float(self.offsetLeI.text()) > float(self.offsetLeO.text())
            and self.offsetsInRange()
        ):
            print("Adding hx")
            if self.rButton.isChecked():
                print("addhxr")
                self.addHxR()
            if self.lButton.isChecked():
                print("addhxl")
                self.addHxL()
        else:
            msgb = QMessageBox()
            msgb.setText(
                "At least 20% of difference and larger top port than bottom port needed and valid range [0, 100]"
            )
            msgb.exec_()

    def minOffsetDistance(self):
        return abs(float(self.offsetLeI.text()) - float(self.offsetLeO.text())) >= 5

    def offsetsInRange(self):
        return (0 <= float(self.offsetLeI.text()) <= 100) and (
            0 <= float(self.offsetLeO.text()) <= 100
        )

    def addHxL(self):
        self._addHeatExchanger(_side.Side.LEFT)

    def addHxR(self):
        self._addHeatExchanger(_side.Side.RIGHT)

    def _addHeatExchanger(self, side: _side.Side):
        name = self.hxNameLe.text()
        if not name:
            messageBox = QMessageBox()
            messageBox.setText(
                "Please specify the name of the heat exchanger that you want to add."
            )
            messageBox.exec_()
            return

        relativeInputHeight = float(self.offsetLeI.text()) / 100
        relativeOutputHeight = float(self.offsetLeO.text()) / 100

        heatExchanger = self.storage.addHeatExchanger(name, side, relativeInputHeight, relativeOutputHeight)

        itemText = self._getHeatExchangerListItemText(heatExchanger)
        if side == _side.Side.LEFT:
            self.leftHeatExchangersItemListWidget.addItem(itemText)
        else:
            self.rightHeatExchangersItemListWidget.addItem(itemText)

    def manAddPortPair(self):
        if float(self.manPortLeI.text()) > 100:
            self.manPortLeI.setText("100")

        if float(self.manPortLeO.text()) < 0:
            self.manPortLeO.setText("0")

        self.storage.addDirectPortPair(
            self.manlButton.isChecked(),
            float(self.manPortLeI.text()) / 100,
            float(self.manPortLeO.text()) / 100,
            self.storage.h,
        )

        self._leftDirectPortPairsItemListWidget.clear()
        self._rightDirectPortPairsItemListWidget.clear()
        self._loadDirectPortPairs()

    def manRemovePortPairLeft(self):
        self._removeSelectedPortPairs(self._leftDirectPortPairsItemListWidget)

    def manRemovePortPairRight(self):
        self._removeSelectedPortPairs(self._rightDirectPortPairsItemListWidget)

    def _removeSelectedPortPairs(self, directPortPairsListWidget):
        for selectedItem in directPortPairsListWidget.selectedItems():
            selectedDirectPortPair = selectedItem.data(_qtc.Qt.UserRole)

            self.storage.directPortPairs.remove(selectedDirectPortPair)

            row = directPortPairsListWidget.row(selectedItem)
            directPortPairsListWidget.takeItem(row)

            while len(selectedDirectPortPair.fromPort.connectionList) > 0:
                selectedDirectPortPair.fromPort.connectionList[0].deleteConn()

            while len(selectedDirectPortPair.toPort.connectionList) > 0:
                selectedDirectPortPair.toPort.connectionList[0].deleteConn()

            self.storage.inputs.remove(selectedDirectPortPair.fromPort)
            self.storage.outputs.remove(selectedDirectPortPair.toPort)

            self.storage.parent.scene().removeItem(selectedDirectPortPair.fromPort)
            self.storage.parent.scene().removeItem(selectedDirectPortPair.toPort)

    def removeHxL(self):
        self._removeSelectedHeatExchangers(self.leftHeatExchangersItemListWidget)

    def removeHxR(self):
        self._removeSelectedHeatExchangers(self.rightHeatExchangersItemListWidget)

    def _removeSelectedHeatExchangers(self, heatExchangersItemListWidget):
        for selectedItem in heatExchangersItemListWidget.selectedItems():
            heatExchanger = selectedItem.data(_qtc.Qt.UserRole)

            self.storage.heatExchangers.remove(heatExchanger)

            row = heatExchangersItemListWidget.row(selectedItem)
            heatExchangersItemListWidget.takeItem(row)

            while len(heatExchanger.port1.connectionList) > 0:
                heatExchanger.port1.connectionList[0].deleteConn()

            while len(heatExchanger.port2.connectionList) > 0:
                heatExchanger.port2.connectionList[0].deleteConn()

            self.storage.inputs.remove(heatExchanger.port1)
            self.storage.outputs.remove(heatExchanger.port2)

            self.storage.parent.scene().removeItem(heatExchanger.port1)
            self.storage.parent.scene().removeItem(heatExchanger.port2)
            self.storage.parent.scene().removeItem(heatExchanger)

    def modifyHx(self):
        """
        Modify Hx.
        """
        result = self._getFirstSelectedItemAndHeatExchanger()
        if not result:
            return
        selectedItem, heatExchanger = result

        modifyDialog = ModifyRelativeHeightsDialog(
            heatExchanger.relativeInputHeight, heatExchanger.relativeOutputHeight
        )
        newHeights = modifyDialog.newRelativeHeights
        if not newHeights:
            return

        newInputHeight = (
            newHeights.input
            if newHeights.input != "empty"
            else heatExchanger.relativeInputHeight
        )
        newOutputHeight = (
            newHeights.output
            if newHeights.output != "empty"
            else heatExchanger.relativeOutputHeight
        )
        heatExchanger.setRelativeHeights(newInputHeight, newOutputHeight)

        listText = self._getHeatExchangerListItemText(heatExchanger)
        selectedItem.setText(listText)

    def _getFirstSelectedItemAndHeatExchanger(self):
        leftSelectedItems = self.leftHeatExchangersItemListWidget.selectedItems()
        rightSelectedItems = self.rightHeatExchangersItemListWidget.selectedItems()

        if leftSelectedItems:
            side = 0
            selectedItem = leftSelectedItems[0]
        elif rightSelectedItems:
            side = 2
            selectedItem = rightSelectedItems[0]
        else:
            return None

        name = selectedItem.text().split(",")[0]
        for heatExchanger in self.storage.heatExchangers:
            if heatExchanger.displayName == name and heatExchanger.sSide == side:
                return selectedItem, heatExchanger

        raise AssertionError(f"No heat exchanger with name {name} found.")

    def modifyPort(self):
        """
        Modify existing ports.
        """
        selectedItem = self._getFirstSelectedDirectPortPairListWidgetItem()
        if not selectedItem:
            return

        directPortPair: _dpp.DirectPortPair = selectedItem.data(_qtc.Qt.UserRole)

        dialogResult = ModifyRelativeHeightsDialog(
            directPortPair.relativeInputHeight, directPortPair.relativeOutputHeight
        )
        newHeights = dialogResult.newRelativeHeights
        if not newHeights:
            return

        newRelativeInputHeight = (
            newHeights.input
            if newHeights.input != "empty"
            else directPortPair.relativeInputHeight
        )
        newRelativeOutputHeight = (
            newHeights.output
            if newHeights.output != "empty"
            else directPortPair.relativeOutputHeight
        )

        directPortPair.setRelativeHeights(
            newRelativeInputHeight, newRelativeOutputHeight, self.storage.h
        )

        newText = self._getDirectPortPairListItemText(directPortPair)
        selectedItem.setText(newText)

    def _getFirstSelectedDirectPortPairListWidgetItem(
        self,
    ) -> _tp.Optional[QListWidgetItem]:
        leftSelectedItems = self._leftDirectPortPairsItemListWidget.selectedItems()
        if leftSelectedItems:
            return leftSelectedItems[0]

        rightSelectedItems = self._rightDirectPortPairsItemListWidget.selectedItems()
        if rightSelectedItems:
            return rightSelectedItems[0]

        return None

    def incrSize(self):
        self._changeSize(self.HEIGHT_INCREMENT, self.WIDTH_INCREMENT)

    def decrSize(self):
        self._changeSize(-self.HEIGHT_INCREMENT, -self.WIDTH_INCREMENT)

    def _changeSize(self, deltaH, deltaW):
        self.storage.updatePortItemPositions(deltaH, deltaW)
        self.storage.h += deltaH
        self.storage.w += deltaW
        self.storage.updateHeatExchangersAfterTankSizeChange()
        self.storage.updateImage()

    def acceptedEdit(self):
        # print("Changing displayName")
        test = self.le.text()
        if self.le.text() == "":
            qmb = QMessageBox()
            qmb.setText("Please set a name for this storage tank.")
            qmb.setStandardButtons(QMessageBox.Ok)
            qmb.setDefaultButton(QMessageBox.Ok)
            qmb.exec()
            return
        self.storage.setName(self.le.text())
        self.close()

    def cancel(self):
        self.close()



# pylint: skip-file
# type: ignore

import trace as _trc
import typing as _tp
import abc as _abc
import logging as _log
import sys as _sys

_TCo = _tp.TypeVar("_TCo")

__all__ = ["createTracer", "TracerBase"]

_logger = _log.getLogger("root")


class TracerBase(_abc.ABC):
    @_abc.abstractmethod
    def run(self, func: _tp.Callable[[], _TCo]) -> _TCo:
        pass


class _DummyTracer(TracerBase):
    def run(self, func: _tp.Callable[[], _TCo]) -> _TCo:
        return func()


class _Tracer(TracerBase):
    def __init__(self):
        ignoredirs = [_sys.prefix]
        self._trace = _trc.Trace(count=False, trace=True, timing=True, ignoredirs=ignoredirs)

    def run(self, func: _tp.Callable[[], _TCo]) -> _TCo:
        return self._trace.runfunc(func)


def createTracer(shallTrace: bool) -> TracerBase:
    if not shallTrace:
        _logger.debug("Tracing is DISABLED.")
        return _DummyTracer()

    _logger.info("Tracing is ENABLED.")
    return _Tracer()



# pylint: skip-file
# type: ignore

import copy as _cp
import dataclasses as _dc
import json as _json
import uuid as _uuid
import typing as _tp

import pytest as _pt

import trnsysGUI.serialization as _ser


@_dc.dataclass
class PersonVersion0(_ser.UpgradableJsonSchemaMixinVersion0):
    firstName: str
    age: int
    heightInM: float

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("ff2ba3c8-4fef-4a64-a026-11212ab35d6b")


@_dc.dataclass
class PersonVersion1(_ser.UpgradableJsonSchemaMixin):
    firstName: str

    lastName: str
    age: int
    heightInCm: int

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[PersonVersion0]:
        return PersonVersion0

    @classmethod
    def upgrade(cls, superseded: PersonVersion0) -> "PersonVersion1":
        lastName = ""
        heightInCm = round(superseded.heightInM * 100)
        return PersonVersion1(
            superseded.firstName, lastName, superseded.age, heightInCm
        )

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("70d5694f-032c-4ca8-b13c-c020b05f2179")


@_dc.dataclass
class HasTitle:
    title: str


@_dc.dataclass
class Person(HasTitle, _ser.UpgradableJsonSchemaMixin):
    lastName: str
    ageInYears: int
    heightInCm: int

    @classmethod
    def getSupersededClass(cls) -> _tp.Type[PersonVersion1]:
        return PersonVersion1

    @classmethod
    def upgrade(cls, superseded: PersonVersion1) -> "Person":
        title = ""
        ageInYears = superseded.age
        return Person(title, superseded.lastName, ageInYears, superseded.heightInCm)

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("1774d088-3917-4c29-a76a-0a4514ef6cf5")


@_dc.dataclass
class Team(_ser.UpgradableJsonSchemaMixinVersion0):
    members: _tp.Sequence[Person]

    def __post_init__(self):
        if not self.members:
            raise ValueError("Team must have at least one member.")

    @classmethod
    def getVersion(cls) -> _uuid.UUID:
        return _uuid.UUID("9e322099-c043-4f23-b6df-4087bb5950d7")


class TestSerialization:
    SERIALIZED_P0 = {
        "__version__": "ff2ba3c8-4fef-4a64-a026-11212ab35d6b",
        "age": 32,
        "firstName": "Damian",
        "heightInM": 1.73,
    }
    SERIALIZED_P1 = {
        "__version__": "70d5694f-032c-4ca8-b13c-c020b05f2179",
        "age": 32,
        "firstName": "Damian",
        "heightInCm": 173,
        "lastName": "Birchler",
    }
    SERIALIZED_P = {
        "__version__": "1774d088-3917-4c29-a76a-0a4514ef6cf5",
        "ageInYears": 32,
        "heightInCm": 173,
        "lastName": "Birchler",
        "title": "Mr.",
    }
    SERIALIZED_T0 = {
        "__version__": "9e322099-c043-4f23-b6df-4087bb5950d7",
        "members": [
            {
                "__version__": "ff2ba3c8-4fef-4a64-a026-11212ab35d6b",
                "age": 32,
                "firstName": "Damian",
                "heightInM": 1.73,
            },
            {
                "__version__": "70d5694f-032c-4ca8-b13c-c020b05f2179",
                "age": 32,
                "firstName": "Damian",
                "heightInCm": 173,
                "lastName": "Birchler",
            },
            {
                "__version__": "1774d088-3917-4c29-a76a-0a4514ef6cf5",
                "ageInYears": 32,
                "heightInCm": 173,
                "lastName": "Birchler",
                "title": "Mr.",
            },
        ],
    }

    def testSerialization(self):
        p0 = PersonVersion0(firstName="Damian", age=32, heightInM=1.73)
        assert p0.to_dict() == self.SERIALIZED_P0

        p1 = PersonVersion1(
            firstName="Damian", lastName="Birchler", age=32, heightInCm=173
        )
        assert p1.to_dict() == self.SERIALIZED_P1

        p = Person(title="Mr.", lastName="Birchler", ageInYears=32, heightInCm=173)
        assert p.to_dict() == self.SERIALIZED_P

        t = Team([p0, p1, p])
        assert t.to_dict() == self.SERIALIZED_T0

    def testStandardUseCase(self):
        json = _json.dumps(self.SERIALIZED_P0)

        p = Person.from_json(json)

        assert p.title is ""
        assert p.lastName is ""
        assert p.ageInYears == self.SERIALIZED_P0["age"]
        assert p.heightInCm == self.SERIALIZED_P0["heightInM"] * 100

    def testWrongVersionRaises(self):
        serializedP1 = _cp.deepcopy(self.SERIALIZED_P1)
        serializedP1["__version__"] = 4
        json = _json.dumps(serializedP1)

        with _pt.raises(_ser.SerializationError):
            Person.from_json(json)

    def testWrongVersion0Raises(self):
        serializedP1 = _cp.deepcopy(self.SERIALIZED_P0)
        serializedP1["__version__"] = "2cba32bd-4c8a-49fc-9f0b-a6b312adcf24"
        json = _json.dumps(serializedP1)

        with _pt.raises(_ser.SerializationError):
            Person.from_json(json)

    def testMissingVersion0DoesNotRaise(self):
        serializedP0 = _cp.deepcopy(self.SERIALIZED_P0)
        del serializedP0["__version__"]

        json = _json.dumps(serializedP0)

        Person.from_json(json)

    def testMissingVersionRaises(self):
        serializedP = _cp.deepcopy(self.SERIALIZED_P)
        del serializedP["__version__"]

        json = _json.dumps(serializedP)

        with _pt.raises(_ser.SerializationError):
            Person.from_json(json)

    def testLoadVersion1(self):
        json = _json.dumps(self.SERIALIZED_P1)

        p1 = PersonVersion1.from_json(json)
        p = Person.from_json(json)

        assert p.title is ""
        assert p.lastName == p1.lastName
        assert p.ageInYears == p1.age
        assert p.heightInCm == p1.heightInCm

    def testWrongFormatRaises(self):
        phonyP1 = self.SERIALIZED_P0.copy()
        firstName = phonyP1["firstName"]
        del phonyP1["firstName"]
        phonyP1["first-mohican"] = firstName

        json = _json.dumps(phonyP1)

        with _pt.raises(_ser.SerializationError):
            Person.from_json(json)

        with _pt.raises(_ser.SerializationError):
            PersonVersion0.from_json(json)

    def testNested(self):
        json = _json.dumps(self.SERIALIZED_T0)

        assert Team.json_schema() == {
            "$schema": "http://json-schema.org/draft-06/schema#",
            "definitions": {
                "Person": {
                    "anyOf": [
                        {
                            "description": "Person(title: str, "
                            "lastName: str, "
                            "ageInYears: int, "
                            "heightInCm: int)",
                            "properties": {
                                "__version__": {
                                    "const": "1774d088-3917-4c29-a76a-0a4514ef6cf5"
                                },
                                "ageInYears": {"type": "integer"},
                                "heightInCm": {"type": "integer"},
                                "lastName": {"type": "string"},
                                "title": {"type": "string"},
                            },
                            "required": [
                                "title",
                                "lastName",
                                "ageInYears",
                                "heightInCm",
                                "__version__",
                            ],
                            "type": "object",
                        },
                        {"$ref": "#/definitions/PersonVersion1"},
                    ]
                },
                "PersonVersion0": {
                    "description": "PersonVersion0(firstName: "
                    "str, age: int, heightInM: "
                    "float)",
                    "properties": {
                        "__version__": {
                            "const": "ff2ba3c8-4fef-4a64-a026-11212ab35d6b"
                        },
                        "age": {"type": "integer"},
                        "firstName": {"type": "string"},
                        "heightInM": {"type": "number"},
                    },
                    "required": ["firstName", "age", "heightInM"],
                    "type": "object",
                },
                "PersonVersion1": {
                    "anyOf": [
                        {
                            "description": "PersonVersion1(firstName: "
                            "str, lastName: "
                            "str, age: int, "
                            "heightInCm: "
                            "int)",
                            "properties": {
                                "__version__": {
                                    "const": "70d5694f-032c-4ca8-b13c-c020b05f2179"
                                },
                                "age": {"type": "integer"},
                                "firstName": {"type": "string"},
                                "heightInCm": {"type": "integer"},
                                "lastName": {"type": "string"},
                            },
                            "required": [
                                "firstName",
                                "lastName",
                                "age",
                                "heightInCm",
                                "__version__",
                            ],
                            "type": "object",
                        },
                        {"$ref": "#/definitions/PersonVersion0"},
                    ]
                },
            },
            "description": "Team(members: Sequence[tests.trnsysGUI.testSerialization.Person])",
            "properties": {
                "__version__": {"const": "9e322099-c043-4f23-b6df-4087bb5950d7"},
                "members": {"items": {"$ref": "#/definitions/Person"}, "type": "array"},
            },
            "required": ["members"],
            "type": "object",
        }
        team = Team.from_json(json)

        expectedMembers = [
            Person.from_dict(d)
            for d in [self.SERIALIZED_P0, self.SERIALIZED_P1, self.SERIALIZED_P]
        ]

        assert team.members == expectedMembers



__all__ = [
    "CreateProject",
    "LoadProject",
    "MigrateProject",
    "Project",
    "getProject",
    "getExistingEmptyDirectory",
    "getLoadOrMigrateProject",
]

import dataclasses as _dc
import enum as _enum
import pathlib as _pl
import typing as _tp

import PyQt5.QtWidgets as _qtw

import trnsysGUI.common.cancelled as _ccl


@_dc.dataclass
class CreateProject:
    jsonFilePath: _pl.Path


@_dc.dataclass
class LoadProject:
    jsonFilePath: _pl.Path


@_dc.dataclass
class MigrateProject:
    oldJsonFilePath: _pl.Path
    newProjectFolderPath: _pl.Path


Project = _tp.Union[CreateProject, LoadProject, MigrateProject]


def getProject() -> _ccl.MaybeCancelled[Project]:
    createOpenMaybeCancelled = _askUserWhetherToCreateNewProjectOrOpenExisting()

    while not _ccl.isCancelled(createOpenMaybeCancelled):
        createOpen = _ccl.value(createOpenMaybeCancelled)

        projectMaybeCancelled = _getProjectInternal(createOpen)
        if not _ccl.isCancelled(projectMaybeCancelled):
            project = _ccl.value(projectMaybeCancelled)
            return _tp.cast(Project, project)  # Don't know why mypy requires this cast

        createOpenMaybeCancelled = _askUserWhetherToCreateNewProjectOrOpenExisting()

    return _ccl.CANCELLED


class _CreateNewOrOpenExisting(_enum.Enum):
    CREATE_NEW = _enum.auto()
    OPEN_EXISTING = _enum.auto()


def _askUserWhetherToCreateNewProjectOrOpenExisting() -> _ccl.MaybeCancelled[
    _CreateNewOrOpenExisting
]:
    messageBox = _qtw.QMessageBox()
    messageBox.setWindowTitle("Start a new or open an existing project")
    messageBox.setText("Do you want to start a new project or open an existing one?")

    createButton = _qtw.QPushButton("New")
    openButton = _qtw.QPushButton("Open")
    messageBox.addButton(createButton, _qtw.QMessageBox.YesRole)
    messageBox.addButton(openButton, _qtw.QMessageBox.NoRole)
    messageBox.addButton(_qtw.QMessageBox.Cancel)
    messageBox.setFocus()
    messageBox.exec()

    clickedButton = messageBox.clickedButton()

    cancelButton = messageBox.button(_qtw.QMessageBox.Cancel)
    if clickedButton is cancelButton:
        return _ccl.CANCELLED

    if clickedButton is createButton:
        return _CreateNewOrOpenExisting.CREATE_NEW

    if clickedButton is openButton:
        return _CreateNewOrOpenExisting.OPEN_EXISTING

    raise AssertionError("Unknown button was clicked.")


def _getProjectInternal(createOrOpenExisting: "_CreateNewOrOpenExisting") -> _ccl.MaybeCancelled[Project]:
    if createOrOpenExisting == _CreateNewOrOpenExisting.OPEN_EXISTING:
        return getLoadOrMigrateProject()

    if createOrOpenExisting == _CreateNewOrOpenExisting.CREATE_NEW:
        return getCreateProject()

    raise AssertionError(f"Unknown value for enum {_CreateNewOrOpenExisting}: {createOrOpenExisting}")


def getCreateProject(startingDirectoryPath: _tp.Optional[_pl.Path] = None) -> _ccl.MaybeCancelled[CreateProject]:
    projectFolderPathMaybeCancelled = getExistingEmptyDirectory(startingDirectoryPath)
    if _ccl.isCancelled(projectFolderPathMaybeCancelled):
        return _ccl.CANCELLED
    projectFolderPath = _ccl.value(projectFolderPathMaybeCancelled)

    jsonFilePath = projectFolderPath / f"{projectFolderPath.name}.json"

    return CreateProject(jsonFilePath)


def getExistingEmptyDirectory(
    startingDirectoryPath: _tp.Optional[_pl.Path] = None,
) -> _ccl.MaybeCancelled[_pl.Path]:
    while True:
        selectedDirectoryPathString = _qtw.QFileDialog.getExistingDirectory(
            caption="Select new project directory", directory=str(startingDirectoryPath)
        )
        if not selectedDirectoryPathString:
            return _ccl.CANCELLED

        selectedDirectoryPath = _pl.Path(selectedDirectoryPathString)

        if _isEmptyDirectory(selectedDirectoryPath):
            return selectedDirectoryPath

        errorMessage = "The new project directory must be empty."

        messageBox = _qtw.QMessageBox()
        messageBox.setText(errorMessage)
        messageBox.exec()


def _isEmptyDirectory(path: _pl.Path) -> bool:
    if not path.is_dir():
        return False

    containedFilesAndDirectories = list(path.iterdir())

    isDirectoryEmpty = len(containedFilesAndDirectories) == 0

    return isDirectoryEmpty


def getLoadOrMigrateProject() -> _ccl.MaybeCancelled[_tp.Union[LoadProject, MigrateProject]]:
    projectFolderPathString, _ = _qtw.QFileDialog.getOpenFileName(
        caption="Open diagram", filter="*.json"
    )
    if not projectFolderPathString:
        return _ccl.CANCELLED
    jsonFilePath = _pl.Path(projectFolderPathString)

    projectFolderPath = jsonFilePath.parent

    containingFolderIsCalledSameAsJsonFile = projectFolderPath.name == jsonFilePath.stem
    ddckFolder = projectFolderPath / "ddck"

    if not containingFolderIsCalledSameAsJsonFile or not ddckFolder.is_dir():
        oldJsonFilePath = jsonFilePath

        messageBox = _qtw.QMessageBox()
        messageBox.setText(
            "The json you are opening does not have a proper project folder environment. "
            "Do you want to continue and create one?"
        )
        messageBox.setStandardButtons(_qtw.QMessageBox.Yes | _qtw.QMessageBox.Cancel)
        messageBox.setDefaultButton(_qtw.QMessageBox.Cancel)
        result = messageBox.exec()
        if result == _qtw.QMessageBox.Cancel:
            return _ccl.CANCELLED

        maybeCancelled = getExistingEmptyDirectory(
            startingDirectoryPath=projectFolderPath.parent
        )
        if _ccl.isCancelled(maybeCancelled):
            return _ccl.CANCELLED
        newProjectFolderPath = _ccl.value(maybeCancelled)

        return MigrateProject(oldJsonFilePath, newProjectFolderPath)

    return LoadProject(jsonFilePath)



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QUndoCommand


class MoveCommand(QUndoCommand):
    def __init__(self, diagramItem, oldPos, descr):
        super(MoveCommand, self).__init__(descr)
        self.oldPos = oldPos
        self.newPos = diagramItem.scenePos()
        self.item = diagramItem

    def redo(self):
        self.item.setPos(self.newPos)

    def undo(self):
        self.item.setPos(self.oldPos)

    # mergeWith does not get execute, as opposed to the Qt reference
    # def mergeWith(self, mc):
    #     print("in mergeWith")
    #     item = mc.item
    #
    #     if self.item != item:
    #         return False
    #
    #     self.newPos = item.scenePos()
    #
    #     return True



# pylint: skip-file
# type: ignore

import json

from trnsysGUI.AirSourceHP import AirSourceHP
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.Boiler import Boiler
from trnsysGUI.Collector import Collector
from trnsysGUI.Connection import Connection
from trnsysGUI.Connector import Connector
from trnsysGUI.Control import Control
from trnsysGUI.ExternalHx import ExternalHx
from trnsysGUI.GenericBlock import GenericBlock
from trnsysGUI.GroundSourceHx import GroundSourceHx
from trnsysGUI.HPDoubleDual import HPDoubleDual
from trnsysGUI.HeatPump import HeatPump
from trnsysGUI.HeatPumpTwoHx import HeatPumpTwoHx
from trnsysGUI.IceStorage import IceStorage
from trnsysGUI.IceStorageTwoHx import IceStorageTwoHx
from trnsysGUI.MasterControl import MasterControl
from trnsysGUI.PV import PV
from trnsysGUI.PitStorage import PitStorage
from trnsysGUI.Pump import Pump
from trnsysGUI.Radiator import Radiator
from trnsysGUI.StorageTank import StorageTank
from trnsysGUI.TVentil import TVentil
from trnsysGUI.TeePiece import TeePiece
from trnsysGUI.WTap import WTap
from trnsysGUI.WTap_main import WTap_main


class DecoderPaste(json.JSONDecoder):
    """
    Decodes the clipboard.
    It creates the copied blocks with suffix COPY, it does not set the ids forward.

    """

    def __init__(self, *args, **kwargs):
        self.editor = kwargs["editor"]
        self.logger = self.editor.logger
        kwargs.pop("editor")
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, arr):
        """
        This is the decoding function from the clipboard. It seems to get executed for every sub dictionary in the json file.
        By looking for the specific key containing the name of dict elements, one can extract the needed dict.
        The name of the dicts is important because the order in which they are loaded matters (some objects depend on others)

        Parameters
        ----------
        arr

        Returns
        -------

        """

        resBlockList = []
        offset_x = 300
        offset_y = 100
        if ".__BlockDct__" in arr:

            resConnList = []
            self.logger.debug("keys are " + str(sorted((arr.keys()))))
            for k in sorted((arr.keys())):
                if type(arr[k]) is dict:
                    if ".__BlockDict__" in arr[k]:
                        i = arr[k]

                        if i["BlockName"] == "TeePiece":
                            bl = TeePiece(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "TVentil":
                            bl = TVentil(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "Pump":
                            bl = Pump(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "Collector":
                            bl = Collector(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "Kollektor":
                            i["BlockName"] = "Collector"
                            bl = Collector(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "HP":
                            bl = HeatPump(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "IceStorage":
                            bl = IceStorage(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "PitStorage":
                            bl = PitStorage(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "Radiator":
                            bl = Radiator(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "WTap":
                            bl = WTap(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "WTap_main":
                            bl = WTap_main(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "Connector":
                            bl = Connector(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "GenericBlock":
                            bl = GenericBlock(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "Boiler":
                            bl = Boiler(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "AirSourceHP":
                            bl = AirSourceHP(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "PV":
                            bl = PV(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "GroundSourceHx":
                            bl = GroundSourceHx(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )

                        # [--- New encoding
                        elif i["BlockName"] == "StorageTank":
                            bl = StorageTank(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "HeatPump":
                            bl = HeatPump(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "HPTwoHx":
                            bl = HeatPumpTwoHx(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "HPDoubleDual":
                            bl = HPDoubleDual(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "ExternalHx":
                            bl = ExternalHx(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "IceStorageTwoHx":
                            bl = IceStorageTwoHx(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "GenericBlock":
                            bl = GenericBlock(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "MasterControl":
                            bl = MasterControl(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        elif i["BlockName"] == "Control":
                            bl = Control(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loaded=True
                            )
                        # new encoding ---]

                        else:
                            bl = BlockItem(
                                i["BlockName"], self.editor.diagramView, displayName=i["BlockName"], loaded=True
                            )

                        bl.decodePaste(i, offset_x, offset_y, resConnList, resBlockList)

                    elif ".__ConnectionDict__" in arr[k]:
                        # It is important to load the connections at last because else the ports are not defined.
                        i = arr[k]

                        fport = None
                        tPort = None
                        self.logger.debug("Loading a connection in paste")
                        self.logger.debug("blocks are " + str(resBlockList))

                        for connBl in resBlockList:
                            # if "COPY" in connBl.displayName and connBl.displayName[-4:None] == 'COPY':
                            if True:
                                for p in connBl.inputs + connBl.outputs:
                                    if p.id == i["PortFromID"]:
                                        fport = p
                                    if p.id == i["PortToID"]:
                                        tPort = p

                        if fport is None:
                            self.logger.debug("Did not found a fromPort")

                        if tPort is None:
                            self.logger.debug("Did not found a tPort")

                        if True:
                            for cornerL in i["CornerPositions"]:
                                cornerL[0] += offset_x
                                cornerL[1] += offset_y

                            c = Connection(
                                fport,
                                tPort,
                                self.editor,
                                fromPortId=i["PortFromID"],
                                toPortId=i["PortToID"],
                                segmentsLoad=i["SegmentPositions"],
                                cornersLoad=i["CornerPositions"],
                                loadedConn=True,
                            )
                            c.setName(i["ConnDisplayName"])

                            # Note: This wouldn't allow two connections to the same port (which is not really used, but ok)
                            # fport.id = getID()
                            # tPort.id = getID()
                            resConnList.append(c)
                        else:
                            self.logger.debug(
                                "This is an internal connection (e.g. in the storage) and thus is not created now"
                            )

                    elif "__idDct__" in arr[k]:
                        resBlockList.append(arr[k])
                    else:
                        self.logger.debug("Error: Not recognized object in decoder")

            return resBlockList, resConnList

        return arr



# pylint: skip-file
# type: ignore

from PyQt5.QtCore import QPointF, QRectF, Qt
from PyQt5.QtGui import QColor, QPen
from PyQt5.QtWidgets import QGraphicsScene, QGraphicsRectItem
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.Connection import Connection
from trnsysGUI.Graphicaltem import GraphicalItem
from trnsysGUI.Group import Group
from trnsysGUI.ResizerItem import ResizerItem
from trnsysGUI.StorageTank import StorageTank


class Scene(QGraphicsScene):
    """
    This class serves as container for QGraphicsItems and is used in combination with the View to display the
    diagram.
    It contains a rectangle for copy-paste or selecting multiple items.

    Attributes
    ----------

    sRstart : :obj:`QPointF`
        Upper left corner position of selectionRect
    sRh : int
        selectionRectHeight
    sRw : int
        selectionRectWidth
    selectionRect : :obj:`QGraphicsRectItem`
        Rectangle that displays the selection
    viewRect1 : :obj:`QGraphicsRectItem`
        Used to set the initial Scene size to approximately View size
    viewRect2 : :obj:`QGraphicsRectItem`
            Used to set the initial Scene size to approximately View size
    released : bool
        Enables the display of the selectionRect
    pressed : bool
        Set to True when selectionMode is True and mousePressed
        Set to False when no element in selection

    """

    def __init__(self, parent=None):
        # Static size
        # super(Scene, self).__init__(QRectF(0, 0, parent.height(), 1500), parent)
        # self.setSceneRect(0, 0, parent.height(), parent.width())

        # Dynamic size, but "zero" at beginning
        super(Scene, self).__init__(parent)

        self.logger = parent.logger

        self.sRstart = QPointF(-500, -400)
        self.sRh = 700
        self.sRw = 400

        self.selectionRect = QGraphicsRectItem(self.sRstart.x(), self.sRstart.y(), self.sRw, self.sRw)
        self.viewRect1 = QGraphicsRectItem(0, 0, 10, 10)
        self.viewRect2 = QGraphicsRectItem(-800, -400, 10, 10)
        rectColor = QColor(100, 160, 245)

        p1 = QPen(rectColor, 2)
        self.selectionRect.setPen(p1)
        self.viewRect1.setPen(p1)
        self.viewRect2.setPen(p1)

        self.selectionRect.setVisible(False)
        self.viewRect1.setVisible(False)
        self.viewRect2.setVisible(False)

        self.addItem(self.selectionRect)
        self.addItem(self.viewRect1)
        self.addItem(self.viewRect2)

        self.selectedItem = None

        # self.viewRect1.setPos(-1300, -100)
        # self.viewRect2.setPos(-1300, -100)

        self.released = False
        self.pressed = False

    def mouseMoveEvent(self, mouseEvent):
        self.parent().sceneMouseMoveEvent(mouseEvent)
        super(Scene, self).mouseMoveEvent(mouseEvent)

        if self.parent().selectionMode and not self.released and self.pressed:
            self.selectionRect.setVisible(True)
            self.sRw = mouseEvent.scenePos().x() - self.sRstart.x()
            self.sRh = mouseEvent.scenePos().y() - self.sRstart.y()

            if self.sRw < 0 and self.sRh > 0:  # from top right to bottom left
                rectangleR1 = QRectF(self.sRstart.x(), self.sRstart.y(), -abs(self.sRw), self.sRh).normalized()
                self.selectionRect.setRect(rectangleR1)
            elif self.sRw < 0 and self.sRh < 0:  # from bottom right to top left
                rectangleR1 = QRectF(self.sRstart.x(), self.sRstart.y(), -abs(self.sRw), -abs(self.sRh)).normalized()
                self.selectionRect.setRect(rectangleR1)
            elif self.sRw > 0 and self.sRh < 0:  # from bottom left to top right
                rectangleR1 = QRectF(self.sRstart.x(), self.sRstart.y(), self.sRw, -abs(self.sRh)).normalized()
                self.selectionRect.setRect(rectangleR1)
            else:
                rectangleR1 = QRectF(self.sRstart.x(), self.sRstart.y(), self.sRw, self.sRh)
                self.selectionRect.setRect(rectangleR1)

    def mouseReleaseEvent(self, mouseEvent):
        self.logger.debug("Releasing mouse in Scene...")
        self.parent().sceneMouseReleaseEvent(mouseEvent)
        super(Scene, self).mouseReleaseEvent(mouseEvent)
        self.parent().moveDirectPorts = False
        if self.parent().pasting:
            # Dismantle group
            self.parent().clearCopyGroup()

        if self.parent().itemsSelected:
            # Dismantle selection
            self.parent().clearSelectionGroup()

        if self.parent().selectionMode:
            # self.released = True
            self.logger.debug("There are elements inside the selection : " + str(self.hasElementsInRect()))
            if self.hasElementsInRect():
                if self.parent().groupMode:
                    g = self.createGroup()
                    # groupDlg(g, self.parent(), self.elementsInRect())
                    self.parent().showGroupDlg(g, self.elementsInRect())
                elif self.parent().multipleSelectMode:
                    self.parent().createSelectionGroup(self.elementsInRect())
                elif self.parent().copyMode:
                    self.parent().copyElements()
                else:
                    self.logger.info("No recognized mode for selection")
            else:
                self.parent().copyMode = False
                self.parent().selectionMode = False

            self.released = False
            self.pressed = False
            self.selectionRect.setVisible(False)

    def drawBackground(self, painter, rect):
        # Overwrite drawBackground if snapGrid is True
        if self.parent().snapGrid:
            pen = QPen()
            pen.setWidth(2)
            pen.setCosmetic(True)
            painter.setPen(pen)

            gridSize = self.parent().snapSize

            left = int(rect.left()) - (int(rect.left()) % gridSize)
            top = int(rect.top()) - (int(rect.top()) % gridSize)
            points = []
            for x in range(left, int(rect.height()), gridSize):
                for y in range(top, int(rect.bottom()), gridSize):
                    points.append(QPointF(x, y))

            for x in points:
                painter.drawPoint(x)
        else:
            super(Scene, self).drawBackground(painter, rect)

    def keyPressEvent(self, event):
        pass
        if event.key() == Qt.Key_L:
            self.parent().moveDirectPorts = not self.parent().moveDirectPorts
            self.logger.debug("Changing move bool to " + str(self.parent().moveDirectPorts))

    def mousePressEvent(self, event):
        # self.parent().mousePressEvent(event)
        # TODO : remove resizer when click on other block items
        super(Scene, self).mousePressEvent(event)

        if len(self.items(event.scenePos())) > 0:
            self.selectedItem = self.items(event.scenePos())
            # for items in self.items():
            #     if items != self.selectedItem[0]:
            #         if isinstance(items, ResizerItem):
            #             break
            #         else:
            #             if hasattr(items, 'resizer'):
            #                 self.removeItem(items.resizer)
            #                 items.deleteResizer()
            #     else:
            #         print(items)
            #         print(self.selectedItem[0])

        """For selection when clicking on empty space"""
        if len(self.items(event.scenePos())) == 0:
            self.logger.debug("No items here!")
            self.parent().clearSelectionGroup()
            self.parent().selectionMode = True
            self.parent().copyMode = False
            self.parent().groupMode = False
            self.parent().multipleSelectMode = True
            for c in self.parent().connectionList:
                if not self.parent().parent().massFlowEnabled:
                    c.unhighlightConn()

            self.parent().alignYLineItem.setVisible(False)

            for st in self.parent().trnsysObj:
                if isinstance(st, StorageTank):
                    for hx in st.heatExchangers:
                        hx.unhighlightHx()

            if self.selectedItem is not None:
                for items in self.selectedItem:
                    if isinstance(items, GraphicalItem) and hasattr(items, "resizer"):
                        self.removeItem(items.resizer)
                        items.deleteResizer()
                    if isinstance(items, BlockItem) and hasattr(items, "resizer"):
                        self.removeItem(items.resizer)
                        items.deleteResizer()
                    if isinstance(items, ResizerItem):
                        self.removeItem(items)
                        items.parent.deleteResizer()

            if self.selectedItem is not None:
                self.selectedItem.clear()

        # global selectionMode
        if self.parent().selectionMode:
            self.pressed = True
            # self.selectionRect.setParentItem(self)
            if not self.released:
                self.sRstart = event.scenePos()
                self.selectionRect.setRect(
                    self.sRstart.x(),
                    self.sRstart.y(),
                    abs(event.scenePos().x() - self.sRstart.x()),
                    abs(event.scenePos().y() - self.sRstart.y()),
                )
                self.selectionRect.setVisible(True)

        # if len(self.items(event.scenePos())) == 0:
        #     print("No items here!")
        #     for c in self.parent().connectionList:
        #         c.unhighlightConn()
        #
        #     self.parent().alignYLineItem.setVisible(False)
        #
        #     for st in self.parent().trnsysObj:
        #         if isinstance(st, StorageTank):
        #             for hx in st.heatExchangers:
        #                 hx.unhighlightHx()

    def createGroup(self):
        newGroup = Group(self.sRstart.x(), self.sRstart.y(), self.sRw, self.sRh, self)

        return newGroup

    def elementsInRect(self):
        # Return elements in the selection rectangle

        res = []

        for o in self.parent().trnsysObj:
            if isinstance(o, BlockItem):
                self.logger.debug("Checking block to group")
                if self.isInRect(o.scenePos()):
                    res.append(o)

            if type(o) is Connection:
                self.logger.debug("Checking connection to group")
                if self.isInRect(o.fromPort.scenePos()) and self.isInRect(o.toPort.scenePos()):
                    res.append(o)

        for o in self.parent().graphicalObj:
            if isinstance(o, GraphicalItem):
                self.logger.debug("Checking graphic item to group")
                if self.isInRect(o.scenePos()):
                    res.append(o)

        return res

    def hasElementsInRect(self):
        # Check if there are elements in the selection rectangle
        for o in self.parent().trnsysObj:
            if isinstance(o, BlockItem):
                self.logger.debug("Checking block to group")
                if self.isInRect(o.scenePos()):
                    return True

            if type(o) is Connection:
                self.logger.debug("Checking connection to group")
                if self.isInRect(o.fromPort.scenePos()) and self.isInRect(o.toPort.scenePos()):
                    return True

        for o in self.parent().graphicalObj:
            if isinstance(o, GraphicalItem):
                self.logger.debug("Checking graphic item to group")
                if self.isInRect(o.scenePos()):
                    return True

        return False

    def isInRect(self, point):
        # Check if a point is in the selection rectangle
        # from top left to bottom right
        if (
            point.x() > self.sRstart.x()
            and point.x() < (self.sRstart.x() + self.sRw)
            and point.y() > self.sRstart.y()
            and point.y() < (self.sRstart.y() + self.sRh)
        ):
            self.logger.debug("In rect")
            return True

        # from top right to bottom left
        elif (
            point.x() < self.sRstart.x()
            and point.x() > (self.sRstart.x() + self.sRw)
            and point.y() > self.sRstart.y()
            and point.y() < (self.sRstart.y() + self.sRh)
        ):
            self.logger.debug("In rect")
            return True

        # from bottom right to top left
        elif (
            point.x() < self.sRstart.x()
            and point.x() > (self.sRstart.x() + self.sRw)
            and point.y() < self.sRstart.y()
            and point.y() > (self.sRstart.y() + self.sRh)
        ):
            self.logger.debug("In rect")
            return True

        # from bottom left to top right
        elif (
            point.x() > self.sRstart.x()
            and point.x() < (self.sRstart.x() + self.sRw)
            and point.y() < self.sRstart.y()
            and point.y() > (self.sRstart.y() + self.sRh)
        ):
            self.logger.debug("In rect")
            return True

        else:
            return False



# pylint: skip-file
# type: ignore

from PyQt5.QtWidgets import QDialog, QLineEdit, QHBoxLayout, QPushButton, QGridLayout, QLabel


class groupDlg(QDialog):
    def __init__(self, group, parent, itemList):
        super(groupDlg, self).__init__(parent)
        self.setModal(True)
        self.parent = parent
        self.group = group
        self.itemList = itemList
        nameLabel = QLabel("Name:")
        self.le = QLineEdit(self.group.displayName)

        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        buttonLayout = QHBoxLayout()
        buttonLayout.addStretch()
        buttonLayout.addWidget(self.okButton)
        buttonLayout.addWidget(self.cancelButton)
        layout = QGridLayout()
        layout.addWidget(nameLabel, 0, 0)
        layout.addWidget(self.le, 0, 1)
        layout.addLayout(buttonLayout, 1, 0, 2, 0)
        self.setLayout(layout)

        self.okButton.clicked.connect(self.acceptedEdit)
        self.cancelButton.clicked.connect(self.cancel)
        self.setWindowTitle("Set new group")
        self.show()

    def acceptedEdit(self):
        if self.le.text() is not "":
            self.group.setName(self.le.text())
            self.group.setItemsGroup(self.itemList)

            self.parent.selectionMode = False
            self.parent.groupMode = False
            self.close()

    def cancel(self):
        self.parent.selectionMode = False

        self.close()



# pylint: skip-file
# type: ignore

import os
from datetime import datetime
import sys

from PyQt5.QtCore import QPointF
from PyQt5.QtWidgets import (
    QLabel,
    QLineEdit,
    QGridLayout,
    QHBoxLayout,
    QListWidget,
    QPushButton,
    QSpacerItem,
    QVBoxLayout,
    QRadioButton,
    QDialog,
    QTabWidget,
    QWidget,
    QMessageBox,
)


class DifferenceDlg(QDialog):
    """
    This is the dialog box that displays the differences between the export
    file and the reference file.

    exportList and referenceList contains the list of unmatched strings from the
    exported file and reference file respectively.

    fileName refers to the name of the file where the differences are found.
    """

    def __init__(self, parent, exportList, referenceList, fileName):
        super(DifferenceDlg, self).__init__(parent)
        self.parent = parent
        self.fileName = fileName
        self.updatedLines = []

        spacerHeight = 15

        self.tabs = QTabWidget()
        self.tab1 = QWidget()

        self.tabs.addTab(self.tab1, "Export vs Reference")

        description = QLabel("The differences between the files: %s" % fileName)

        g1 = QGridLayout()
        reasonLabel = QLabel("<b>Reason <b>")
        g1.addWidget(reasonLabel, 0, 0, 1, 2)

        qhbL = QHBoxLayout()

        self.listWL = QListWidget()
        for items in exportList:
            # splitedString = items.split(':')
            # lineNo = splitedString[0]
            # eString = splitedString[1]
            # rString = 'Line no: ' + lineNo + ', ' + 'Error String: ' + eString
            self.listWL.addItem(items)
        qhbL.addWidget(self.listWL)

        self.listWR = QListWidget()
        for items in referenceList:
            # splitedString = items.split(',')
            # lineNo = splitedString[0]
            # eString = splitedString[1]
            # rString = 'Line no: ' + lineNo + ', ' + 'Error String: ' + eString
            self.listWR.addItem(items)
        qhbL.addWidget(self.listWR)

        promptLabel = QLabel("Please state reason for change.")
        inputLabel = QLabel("Input reason here:")
        self.reasonInput = QLineEdit()
        self.reasonInput.setPlaceholderText("")

        changeButton = QPushButton("Change")
        changeButton.clicked.connect(self.acceptChange)
        changeAllbutton = QPushButton("Change All")
        changeAllbutton.clicked.connect(self.acceptChangeAll)
        # ignoreButton = QPushButton("Ignore")
        # ignoreButton.clicked.connect(self.ignoreChange)

        g1.addWidget(promptLabel, 1, 0, 1, 1)
        g1.addWidget(inputLabel, 2, 0, 1, 1)
        g1.addWidget(self.reasonInput, 3, 0, 1, 3)
        g1.addWidget(changeButton, 4, 0, 1, 1)
        g1.addWidget(changeAllbutton, 4, 2, 1, 1)
        # g1.addWidget(ignoreButton, 4, 2, 1, 1)
        spaceHx = QSpacerItem(self.width(), spacerHeight)
        g1.addItem(spaceHx, 5, 0, 1, 2)

        self.FinishButton = QPushButton("Finish")
        self.FinishButton.clicked.connect(self.finishEdit)

        l = QVBoxLayout()
        l.addWidget(description)
        l.addWidget(reasonLabel)

        t1Layout = QVBoxLayout()
        t1Layout.addLayout(qhbL)
        t1Layout.addLayout(g1)

        self.tab1.setLayout(t1Layout)

        l.addWidget(self.tabs)
        l.addWidget(self.FinishButton)

        self.setLayout(l)
        self.setFixedSize(1000, 1000)

        self.listWR.setSelectionMode(1)
        self.listWL.setSelectionMode(1)
        self.listWR.clicked.connect(self.listWRClicked)
        self.listWL.clicked.connect(self.listWLClicked)

        self.exec_()

    def listWLClicked(self):
        self.listWR.clearSelection()

    def listWRClicked(self):
        self.listWL.clearSelection()

    def acceptChange(self):
        if len(self.listWR.selectedItems()) > 0:
            msg = QMessageBox()
            msg.setText("Please select from the left side list.")
            msg.exec_()

        if len(self.listWL.selectedItems()) > 0:
            for items in self.listWL.selectedItems():
                # print(items.text()[:items.text().find(":")])
                # print(items.text()[items.text().find(":")+1:])
                lineNo = items.text()[: items.text().find(":")]
                string = items.text()[items.text().find(":") + 1 :]
                self.updatedLines.append(items.text())

        self.updateReferenceFile(lineNo, string)
        self.updateChangelog()
        msg2 = QMessageBox()
        msg2.setText("Changes updated in the reference file.")
        msg2.exec_()

    def acceptChangeAll(self):

        items = []
        lineNoList = []
        stringList = []
        i = 0
        while i < self.listWL.count():
            items.append(self.listWL.item(i))
            i += 1

        for item in items:
            lineNo = item.text()[: item.text().find(":")]
            string = item.text()[item.text().find(":") + 1 :]
            lineNoList.append(lineNo)
            stringList.append(string)
            self.updatedLines.append(item.text())
            # self.updateReferenceFile(lineNo, string)
        self.updateAllChanges(lineNoList, stringList)

        self.updateChangelog()
        msg = QMessageBox()
        msg.setText("All changes updated in the reference file.")
        msg.exec_()

    def finishEdit(self):
        self.close()

    def updateReferenceFile(self, lineNo, string):
        """

        Parameters
        ----------
        lineNo : the line number of the unmatch string
        string : the correct string

        1.Access the reference folder
        2.read from the error file
        3.Update the error lines
        4.write to the error file

        Returns
        -------

        """

        if getattr(sys, "frozen", False):
            ROOT_DIR = os.path.dirname(sys.executable)
        elif __file__:
            ROOT_DIR = os.path.dirname(__file__)
        originalFilePath = os.path.join(ROOT_DIR, "Reference")
        fileToUpdate = os.path.join(originalFilePath, self.fileName)

        with open(fileToUpdate, "r") as file:
            lines = file.readlines()
            # print(lines)

        stringToAdd = string

        try:
            lines[int(lineNo) - 1]
        except IndexError:
            lines.append("")

        lines[int(lineNo) - 1] = stringToAdd

        with open(fileToUpdate, "w") as file:
            file.writelines(lines)

        print("finish running")

    def updateAllChanges(self, lineNoList, stringList):
        """

        Parameters
        ----------
        lineNo : the line number of the unmatch string
        string : the correct string

        1.Access the reference folder
        2.read from the error file
        3.Update the error lines
        4.write to the error file

        Returns
        -------

        """

        if getattr(sys, "frozen", False):
            ROOT_DIR = os.path.dirname(sys.executable)
        elif __file__:
            ROOT_DIR = os.path.dirname(__file__)
        originalFilePath = os.path.join(ROOT_DIR, "Reference")
        fileToUpdate = os.path.join(originalFilePath, self.fileName)

        with open(fileToUpdate, "r") as file:
            lines = file.readlines()

        i = 0
        while i < len(lineNoList):
            try:
                lines[int(lineNoList[i]) - 1]
            except IndexError:
                lines.append("")

            lines[int(lineNoList[i]) - 1] = stringList[i]
            i += 1

        with open(fileToUpdate, "w") as file:
            file.writelines(lines)

        print("finish running")

    def updateChangelog(self):
        linesToAppend = []

        if getattr(sys, "frozen", False):
            ROOT_DIR = os.path.dirname(sys.executable)
        elif __file__:
            ROOT_DIR = os.path.dirname(__file__)
        changelogfile = os.path.join(ROOT_DIR, "changelogs.txt")

        reason = self.reasonInput.text()
        dateAndTime = datetime.now()
        dt_string = dateAndTime.strftime("%d/%m/%Y %H:%M:%S")
        Divider = "==============================================================\n"
        fileName = self.fileName
        updatedLines = self.updatedLines
        reasonHeader = "Reason for change:"
        differenceHeader = "Differences:"
        fileNameHeader = "Name of file:"

        linesToAppend.append(Divider)
        linesToAppend.append(dt_string + "\n" + "\n")
        linesToAppend.append(fileNameHeader + "\n")
        linesToAppend.append(fileName + "\n" + "\n")
        linesToAppend.append(differenceHeader + "\n")
        for lines in updatedLines:
            linesToAppend.append(lines)
        linesToAppend.append("\n" + reasonHeader + "\n")
        linesToAppend.append(reason + "\n")
        linesToAppend.append(Divider)

        with open(changelogfile, "a") as file:
            file.writelines(linesToAppend)



import pathlib as _pl

import trnsysGUI.storageTank.model as _stm


class TestModel:
    def testConvertFromOldestToNewestFormat(self):  # pylint: disable=no-self-use
        expectedPath = (
            _pl.Path(__file__).parent.parent
            / "data"
            / "storageTankConvertedFromOldestToNewestBeforeSettingInvalidIDs.json"
        )
        expectedStorageTankJson = expectedPath.read_text()

        storageTankLegacyJsonPath = _pl.Path(__file__).parent.parent / "data" / "storageTankOldestFormat.json"
        storageTank = _stm.StorageTank.from_json(storageTankLegacyJsonPath.read_text())
        actualStorageTankJson = storageTank.to_json(indent=4, sort_keys=True)

        assert actualStorageTankJson == expectedStorageTankJson

    def testRoundTripWithNewestFormat(self):  # pylint: disable=no-self-use
        newestFormatPath = _pl.Path(__file__).parent.parent / "data" / "storageTankNewestFormat.json"
        newestFormatJson = newestFormatPath.read_text()

        storageTank = _stm.StorageTank.from_json(newestFormatJson)
        actualJson = storageTank.to_json(indent=4, sort_keys=True)

        assert actualJson == newestFormatJson



# pylint: skip-file
# type: ignore

from math import sqrt
import typing as tp

from PyQt5 import QtGui, QtCore
from PyQt5.QtCore import QPointF, QLineF
from PyQt5.QtGui import QColor, QLinearGradient, QBrush, QPen
from PyQt5.QtWidgets import QGraphicsLineItem, QGraphicsTextItem, QMenu

from trnsysGUI.CornerItem import CornerItem
from trnsysGUI.GroupChooserConnDlg import GroupChooserConnDlg
from trnsysGUI.HorizSegmentMoveCommand import HorizSegmentMoveCommand

# This is needed to avoid a circular import but still be able to type check
if tp.TYPE_CHECKING:
    from trnsysGUI.Connection import Connection


def calcDist(p1, p2):
    vec = p1 - p2
    norm = sqrt(vec.x() ** 2 + vec.y() ** 2)
    return norm


class segmentItem(QGraphicsLineItem):
    def __init__(self, startNode, endNode, parent: "Connection"):
        """
        A connection is displayed as a chain of segmentItems (stored in Connection.segments)
        Parameters.
        ----------
        startNode
        endNode
        parent: type(parent): Connection
        """

        super(segmentItem, self).__init__(None)

        self.logger = parent.logger

        self.setFlag(self.ItemIsSelectable, True)

        self.dragged = False
        self.initialised = False
        self.parent = parent

        self.firstChild = None
        self.secondChild = None
        self.cornerChild = None

        self.startNode = startNode
        self.endNode = endNode

        # These nodes are the nodes before and after the crossing
        self.start = None
        self.end = None

        # Unused. Related to interrupting segments for a clearer diagram
        self.disrBeforeNode = None
        self.disrAfterNode = None
        self.disrBeforeSeg = None
        self.disrAfterSeg = None
        self.disrBefore = False
        self.disrAfter = False
        self.hasBridge = False
        self.bridgedSegment = None

        # Only for editorMode 1
        self.firstLine = None
        self.secondLine = None
        self.secondCorner = None
        self.thirdCorner = None

        self.keyPr = 0
        # Used to only create the child objects once
        self._isDraggingInProgress = False

        self.linearGrad = None

        self.insertInParentSegments()
        self.initGrad()

        self.label = QGraphicsTextItem(self.parent.displayName, self.parent.fromPort)
        self.label.setVisible(False)
        self.label.setFlag(self.ItemIsMovable, True)

        self.labelMass = QGraphicsTextItem(self.parent.fromPort)
        self.labelMass.setVisible(False)
        self.labelMass.setFlag(self.ItemIsMovable, True)

        self.setToolTip(self.parent.displayName)

    def segLength(self):
        return calcDist(self.line().p1(), self.line().p2())

    def interpolate(
        self,
        partLen2,
        totLenConn,
    ):
        # c1_r = 0
        # c1_b = 255
        c1_r = 160
        c1_b = 160
        c1_g = 160

        # c2_r = 255
        # c2_b = 0
        c2_r = 0
        c2_b = 0
        c2_g = 0

        try:
            f1 = partLen2 / totLenConn
            f2 = (totLenConn - partLen2) / totLenConn
        except ZeroDivisionError:
            return QColor(100, 100, 100)
        else:
            # return QColor(f1 * c2_r + f2 * c1_r, 0, f1 * c2_b + f2 * c1_b)
            return QColor(f1 * c2_r + f2 * c1_r, f1 * c2_g + f2 * c1_g, f1 * c2_b + f2 * c1_b)

    def initGrad(self):
        """
        Initialize gradient
        Returns
        -------

        """
        # color = QColor(177, 202, 211)
        # color = QColor(3, 124, 193)
        color = QtCore.Qt.red

        pen1 = QtGui.QPen(color, 4)

        # TODO: Dont forget that disr segments can also have parent type not CornerItem
        if type(self.startNode.parent) is CornerItem:
            x1 = self.startNode.firstNode().parent
        else:
            x1 = self.startNode.parent

        if type(self.endNode.parent) is CornerItem:
            x2 = self.endNode.lastNode().parent
        else:
            x2 = self.endNode.parent

        # At init
        self.linearGrad = QLinearGradient(
            QPointF(x1.fromPort.scenePos().x(), x1.fromPort.scenePos().y()),
            QPointF(x2.toPort.scenePos().x(), x2.toPort.scenePos().y()),
        )
        self.linearGrad.setColorAt(0, QtCore.Qt.blue)
        self.linearGrad.setColorAt(1, QtCore.Qt.red)

        self.linearGrad.setColorAt(0, QtCore.Qt.gray)
        self.linearGrad.setColorAt(1, QtCore.Qt.black)

        # self.setPen(QtGui.QPen(color, 2))

        pen1.setBrush(QBrush(self.linearGrad))

        self.setPen(pen1)

    def updateGrad(self):
        """
        Updates the gradient by calling the interpolation function
        Returns
        -------

        """
        # This color is overwritten by the gradient
        color = QtCore.Qt.red
        pen1 = QtGui.QPen(color, 4)

        totLenConn = self.parent.totalLength()
        partLen1 = self.parent.partialLength(self.startNode)
        partLen2 = self.parent.partialLength(self.endNode)

        # self.logger.debug("totlenconn is " + str(totLenConn))
        # self.logger.debug("partlen1 is " + str(partLen1) + "(node1)" + str(self.startNode))
        # self.logger.debug("partlen2  is " + str(partLen2) + "(node2)" + str(self.endNode) +"\n")

        # TODO: Dont forget that disr segments can also have parent type not CornerItem

        if type(self.startNode.parent) is CornerItem:
            # self.logger.debug("In updategrad, startnode parent is corner")
            # self.logger.debug("I have a corner parent, self is " + str(self))

            startGradP = QPointF(self.startNode.parent.scenePos().x(), self.startNode.parent.scenePos().y())
        elif self.startNode.prevN() is None:
            startGradP = QPointF(
                self.startNode.parent.fromPort.scenePos().x(), self.startNode.parent.fromPort.scenePos().y()
            )
        else:
            startGradP = QPointF(self.line().p1().x(), self.line().p1().y())

        if type(self.endNode.parent) is CornerItem:
            # self.logger.debug("In update grad, endnode parent is corner")
            # self.logger.debug("I have a corner parent, self is " + str(self))
            endGradP = QPointF(self.endNode.parent.scenePos().x(), self.endNode.parent.scenePos().y())
        elif self.endNode.nextN() is None:
            endGradP = QPointF(self.endNode.parent.toPort.scenePos().x(), self.endNode.parent.toPort.scenePos().y())
        else:
            endGradP = QPointF(self.line().p2().x(), self.line().p2().y())

        self.linearGrad = QLinearGradient(startGradP, endGradP)

        self.linearGrad.setColorAt(0, self.interpolate(partLen1, totLenConn))
        self.linearGrad.setColorAt(1, self.interpolate(partLen2, totLenConn))

        pen1.setBrush(QBrush(self.linearGrad))

        self.setPen(pen1)

    def insertInParentSegments(self):
        """
        This function inserts the segment in correct order to the segment list of the connection.
        Returns
        -------

        """

        prevSeg = None
        for s in self.parent.segments:
            if s.endNode is self.startNode:
                prevSeg = s

        # Todo: Add support for disr segments

        # if the startNode parent is a connection:
        if not hasattr(self.startNode.parent, "fromPort"):
            self.parent.segments.insert(self.parent.segments.index(prevSeg) + 1, self)
        else:
            self.parent.segments.insert(0, self)

    def mousePressEvent(self, e):

        if e.button() == 1:
            self.keyPr = 1
            self.logger.debug("Setting key to 1")

            self.parent.highlightConn()

            if self.isVertical():
                try:
                    self.oldX = self.startNode.parent.scenePos().x()
                except AttributeError:
                    pass
                else:
                    self.logger.debug("set oldx")

    def mouseMoveEvent(self, e):
        # self.logger.debug("mouse moved")
        # self.logger.debug(str(e.button()))

        self.logger.debug(self.parent.parent.editorMode)
        if self.keyPr == 1:
            self.logger.debug("moved with button 1")
            newPos = e.pos()

            if self.parent.parent.editorMode == 0:
                if not self._isDraggingInProgress:
                    self.initInMode0()
                else:
                    self.dragInMode0(newPos)

            elif self.parent.parent.editorMode == 1:
                # if self.parent.segments[0].isVertical() == False and self.parent.segments[2].isVertical() == False:
                # self.logger.debug(len(self.parent.segments))
                if type(self.startNode.parent) is CornerItem and type(self.endNode.parent) is CornerItem:
                    if not self.startNode.parent.isVisible():
                        self.startNode.parent.setVisible(True)
                    if not self.endNode.parent.isVisible():
                        self.endNode.parent.setVisible(True)
                    if self.isVertical():
                        self.logger.debug("Segment is vertical: %s", self.parent.segments.index(self))
                        self.endNode.parent.setPos(newPos.x(), self.endNode.parent.scenePos().y())
                        self.startNode.parent.setPos(newPos.x(), self.startNode.parent.scenePos().y())
                        self.updateGrad()

                    if self.isHorizontal():
                        self.logger.debug("Segment is horizontal")
                        self.endNode.parent.setPos(self.endNode.parent.scenePos().x(), newPos.y())
                        self.startNode.parent.setPos(self.startNode.parent.scenePos().x(), newPos.y())

                elif type(self.endNode.parent) is CornerItem and self.isVertical():
                    self.logger.debug("Segment is vertical and can't be moved")

                if self.isHorizontal():
                    isFirstSegment = hasattr(self.startNode.parent, "fromPort") and not self.startNode.prevN()
                    isLastSegment = hasattr(self.endNode.parent, "fromPort") and not self.endNode.nextN()

                    if isLastSegment:
                        self.logger.debug("A last segment is being dragged.")
                        if not self._isDraggingInProgress:
                            self._initInMode1(False)
                        self._dragInMode1(False, newPos)
                    elif isFirstSegment:
                        self.logger.debug("A first segment is being dragged.")
                        if not self._isDraggingInProgress:
                            self._initInMode1(True)
                        self._dragInMode1(True, newPos)
            else:
                self.logger.debug("Unrecognized editorMode in segmentItem mouseMoveEvent")

    def mouseDoubleClickEvent(self, event):
        # self.parent.deleteConn()
        return

    def deleteNextHorizSeg(self, b, nextS):
        if b:
            pass
        else:
            nodeTodelete1 = self.endNode
            nodeTodelete2 = self.endNode.nextN()
            self.endNode = nextS.endNode

            self.startNode.setNext(self.endNode)
            self.endNode.setPrev(self.startNode)

            # x-position of the ending point of the next segment line
            posx1 = self.parent.segments[self.parent.segments.index(self) + 2].line().p2().x()

            self.parent.parent.diagramScene.removeItem(nextS)
            self.parent.segments.remove(nextS)
            self.parent.parent.diagramScene.removeItem(nodeTodelete1.parent)

            indexOfSelf = self.parent.segments.index(self)
            nextVS = self.parent.segments[indexOfSelf + 1]

            self.parent.parent.diagramScene.removeItem(nextVS)
            self.parent.segments.remove(nextVS)
            self.parent.parent.diagramScene.removeItem(nodeTodelete2.parent)

            self.setLine(
                self.startNode.parent.scenePos().x(),
                self.startNode.parent.scenePos().y(),
                posx1,
                self.startNode.parent.scenePos().y(),
            )

    def deletePrevHorizSeg(self, b, prevS):
        if b:
            pass
        else:
            nodeTodelete1 = self.startNode
            nodeTodelete2 = self.startNode.prevN()
            self.startNode = prevS.startNode

            self.startNode.setNext(self.endNode)
            self.endNode.setPrev(self.startNode)

            posx1 = self.parent.segments[self.parent.segments.index(self) - 2].line().p1().x()

            self.parent.parent.diagramScene.removeItem(prevS)
            self.parent.segments.remove(prevS)
            self.parent.parent.diagramScene.removeItem(nodeTodelete1.parent)

            indexOfSelf = self.parent.segments.index(self)
            nextVS = self.parent.segments[indexOfSelf - 1]

            self.parent.parent.diagramScene.removeItem(nextVS)
            self.parent.segments.remove(nextVS)
            self.parent.parent.diagramScene.removeItem(nodeTodelete2.parent)

            self.setLine(
                posx1,
                self.endNode.parent.scenePos().y(),
                self.endNode.parent.scenePos().x(),
                self.endNode.parent.scenePos().y(),
            )

    def deleteSegment(self):
        nodeToConnect = self.startNode.prevN()
        nodeToConnect2 = self.endNode.nextN()

        nodeToConnect.setNext(nodeToConnect2)

        self.parent.parent.diagramScene.removeItem(self)
        self.parent.segments.remove(self)
        self.parent.parent.diagramScene.removeItem(self.startNode.parent)
        self.parent.parent.diagramScene.removeItem(self.endNode.parent)

    def splitSegment(self):
        pass

    def mouseReleaseEvent(self, e):
        # Should be same as below
        # self.scene().removeItem(self)

        if e.button() == 1:
            self.keyPr = 0

            if self.parent.parent.editorMode == 0:
                if self._isDraggingInProgress:
                    self.cornerChild.setFlag(self.ItemSendsScenePositionChanges, True)

                    self.hide()
                    self.parent.segments.remove(self)
                    self.parent.parent.diagramScene.removeItem(self)

            elif self.parent.parent.editorMode == 1:
                # if self.parent.segments[0].isVertical() == False and self.parent.segments[2].isVertical() == False:
                if self.isVertical():
                    try:
                        self.oldX
                    except AttributeError:
                        pass
                    else:
                        command = HorizSegmentMoveCommand(self, self.oldX, "Moving segment command")
                        self.parent.parent.parent().undoStack.push(command)
                        self.oldX = self.scenePos().x()

                if self.isHorizontal():
                    if type(self.startNode.parent) is CornerItem and type(self.endNode.parent) is CornerItem:
                        try:
                            nextHorizSeg = self.parent.segments[self.parent.segments.index(self) + 2]
                            prevHorizSeg = self.parent.segments[self.parent.segments.index(self) - 2]
                        except IndexError:
                            self.logger.debug("no next or prev segments")
                        else:
                            # if nextHorizSeg.isHorizontal() and int(nextHorizSeg.line().p2().y()) == int(
                            #         self.endNode.parent.pos().y()): # TODO : Edit here to combine segment
                            # self.logger.debug("Next h seg could be deleted")
                            if nextHorizSeg.isHorizontal() and int(self.endNode.parent.pos().y() - 10) <= int(
                                nextHorizSeg.line().p2().y()
                            ) <= int(self.endNode.parent.pos().y() + 10):
                                self.deleteNextHorizSeg(False, nextHorizSeg)
                                self.logger.debug("next horizontal")
                                return

                            if prevHorizSeg.isHorizontal() and int(self.startNode.parent.pos().y() - 10) <= int(
                                prevHorizSeg.line().p2().y()
                            ) <= int(self.startNode.parent.pos().y() + 10):
                                # self.logger.debug("Prev h seg could be deleted")
                                self.deletePrevHorizSeg(False, prevHorizSeg)
                                self.logger.debug("previous horizontal")
                                return

                if self.secondCorner is not None:
                    self.logger.debug("Second corner is not none")
                    # if PortItem
                    if hasattr(self.endNode.parent, "fromPort"):
                        # self.hide()
                        # self.parent.segments.remove(self)
                        # self.parent.parent.diagramScene.removeItem(self)

                        segbef = self.parent.segments[self.parent.getNodePos(self.secondCorner.node.prevN().parent)]

                        segbef.setLine(
                            segbef.line().p1().x(),
                            segbef.line().p1().y(),
                            segbef.line().p2().x(),
                            self.secondCorner.scenePos().y(),
                        )
                        self.setLine(
                            self.thirdCorner.scenePos().x(),
                            self.thirdCorner.scenePos().y(),
                            self.line().p2().x(),
                            self.line().p2().y(),
                        )
                        self.secondCorner.setFlag(self.ItemSendsScenePositionChanges, True)
                        self.thirdCorner.setFlag(self.ItemSendsScenePositionChanges, True)

                        # Allow for iterative branching
                        self.secondCorner = None
                        self.thirdCorner = None
                        self.firstLine = None
                        self.secondLine = None
                        self._isDraggingInProgress = False

                    # if PortItem
                    elif hasattr(self.startNode.parent, "fromPort"):
                        segafter = self.parent.segments[self.parent.getNodePos(self.thirdCorner.node.nextN().parent)]

                        segafter.setLine(
                            segafter.line().p1().x(),
                            self.thirdCorner.scenePos().y(),
                            segafter.line().p2().x(),
                            segafter.line().p2().y(),
                        )
                        self.setLine(
                            self.line().p1().x(),
                            self.line().p1().y(),
                            self.secondCorner.scenePos().x(),
                            self.secondCorner.scenePos().y(),
                        )

                        self.secondCorner.setFlag(self.ItemSendsScenePositionChanges, True)
                        self.thirdCorner.setFlag(self.ItemSendsScenePositionChanges, True)

                        # Allow for iterative branching
                        self.secondCorner = None
                        self.thirdCorner = None
                        self.firstLine = None
                        self.secondLine = None
                        self._isDraggingInProgress = False

                    else:
                        self.logger.debug("getting no start or end")

                else:
                    self.logger.debug("Second corner is none")
            else:
                pass

    def initInMode0(self):
        if (hasattr(self.startNode.parent, "fromPort")) and (self.startNode.prevN() is not None):
            self.disrAfterNode = self.startNode
            self.start = self.startNode.prevN().prevN()

            segments = self.parent.segments
            for s in segments:
                if s.startNode is self.start:
                    self.disrBeforeSeg = s

            self.disrAfterSeg = self
            self.disrBefore = True

        else:
            self.start = self.startNode

        if (hasattr(self.endNode.parent, "fromPort")) and (self.endNode.nextN() is not None):
            self.disrBeforeNode = self.endNode
            self.end = self.endNode.nextN().nextN()

            segments = self.parent.segments
            for s in segments:
                if s.endNode is self.end:
                    self.disrAfterSeg = s

            self.disrBeforeSeg = self
            self.disrAfter = True

        else:
            self.end = self.endNode

        rad = 2

        self.cornerChild = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.start, self.end, self.parent)
        self.firstChild = segmentItem(self.start, self.cornerChild.node, self.parent)
        self.secondChild = segmentItem(self.cornerChild.node, self.end, self.parent)

        self.start.setNext(self.cornerChild.node)
        self.end.setPrev(self.cornerChild.node)

        self.firstChild.setVisible(False)
        self.secondChild.setVisible(False)
        self.cornerChild.setVisible(False)

        self.parent.parent.diagramScene.addItem(self.firstChild)
        self.parent.parent.diagramScene.addItem(self.secondChild)
        self.parent.parent.diagramScene.addItem(self.cornerChild)

        self._isDraggingInProgress = True

    def _initInMode1(self, b):

        rad = 2

        if b:
            if (hasattr(self.startNode.parent, "fromPort")) and (self.startNode.prevN() is None):
                # We are at the toPort.
                # self.end = self.endNode
                # self.start = self.startNode

                self.secondCorner = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self.parent)
                self.thirdCorner = CornerItem(
                    -rad, -rad, 2 * rad, 2 * rad, self.secondCorner.node, self.endNode, self.parent
                )

                self.secondCorner.node.setNext(self.thirdCorner.node)
                self.startNode.setNext(self.secondCorner.node)
                self.endNode.setPrev(self.thirdCorner.node)

                self.endNode = self.secondCorner.node

                self.firstLine = segmentItem(self.secondCorner.node, self.thirdCorner.node, self.parent)
                self.secondLine = segmentItem(self.thirdCorner.node, self.thirdCorner.node.nextN(), self.parent)

                self.secondCorner.setVisible(False)
                self.thirdCorner.setVisible(False)
                self.firstLine.setVisible(False)
                self.secondLine.setVisible(False)
                # self.thirdLine.setVisible(False)

                self.parent.parent.diagramScene.addItem(self.secondCorner)
                self.parent.parent.diagramScene.addItem(self.thirdCorner)
                self.parent.parent.diagramScene.addItem(self.firstLine)
                self.parent.parent.diagramScene.addItem(self.secondLine)
                self.logger.debug("inited")

                self._isDraggingInProgress = True
        else:
            if (hasattr(self.endNode.parent, "fromPort")) and (self.endNode.nextN() is None):
                # We are at the toPort.
                # self.end = self.endNode
                # self.start = self.startNode

                self.secondCorner = CornerItem(-rad, -rad, 2 * rad, 2 * rad, self.startNode, None, self.parent)
                self.thirdCorner = CornerItem(
                    -rad, -rad, 2 * rad, 2 * rad, self.secondCorner.node, self.endNode, self.parent
                )

                self.secondCorner.node.setNext(self.thirdCorner.node)
                self.startNode.setNext(self.secondCorner.node)
                self.endNode.setPrev(self.thirdCorner.node)

                self.startNode = self.thirdCorner.node

                self.firstLine = segmentItem(self.secondCorner.node.prevN(), self.secondCorner.node, self.parent)
                self.secondLine = segmentItem(self.secondCorner.node, self.thirdCorner.node, self.parent)

                self.secondCorner.setVisible(False)
                self.thirdCorner.setVisible(False)
                self.firstLine.setVisible(False)
                self.secondLine.setVisible(False)
                # self.thirdLine.setVisible(False)

                self.parent.parent.diagramScene.addItem(self.secondCorner)
                self.parent.parent.diagramScene.addItem(self.thirdCorner)
                self.parent.parent.diagramScene.addItem(self.firstLine)
                self.parent.parent.diagramScene.addItem(self.secondLine)
                self.logger.debug("inited")

                self._isDraggingInProgress = True

    def isVertical(self):
        return self.line().p1().x() == self.line().p2().x()

    def isHorizontal(self):
        return self.line().p1().y() == self.line().p2().y()

    def dragInMode0(self, newPos):
        p1 = self.line().p1()
        p2 = self.line().p2()

        if len(self.scene().items(newPos)) == 0:
            self.firstChild.setLine(QLineF(p1.x(), p1.y(), newPos.x(), newPos.y()))
            self.secondChild.setLine(QLineF(newPos.x(), newPos.y(), p2.x(), p2.y()))

            self.cornerChild.setPos(newPos)

            self.firstChild.updateGrad()
            self.secondChild.updateGrad()

            # Bring corner to front
            self.cornerChild.setZValue(100)
            self.firstChild.setZValue(1)
            self.secondChild.setZValue(1)

            self.firstChild.setVisible(True)
            self.secondChild.setVisible(True)
            self.cornerChild.setVisible(True)

    def _dragInMode1(self, b, newPos):
        self.logger.debug("after inited")

        if b:
            self.thirdCorner.setPos(newPos.x() - 10, newPos.y())
            self.secondCorner.setPos(newPos.x() - 10, self.parent.fromPort.scenePos().y())
            self.thirdCorner.node.nextN().parent.setY(newPos.y())

            self.firstLine.setLine(
                self.secondCorner.scenePos().x(),
                self.secondCorner.scenePos().y(),
                self.thirdCorner.scenePos().x(),
                newPos.y(),
            )
            self.secondLine.setLine(
                self.thirdCorner.scenePos().x(),
                self.thirdCorner.scenePos().y(),
                self.thirdCorner.node.nextN().parent.scenePos().x(),
                self.thirdCorner.node.nextN().parent.scenePos().y(),
            )
            self.setLine(
                self.startNode.parent.fromPort.scenePos().x(),
                self.startNode.parent.fromPort.scenePos().y(),
                self.secondCorner.scenePos().x(),
                self.secondCorner.scenePos().y(),
            )

            self.secondCorner.setZValue(100)
            self.thirdCorner.setZValue(100)
            self.firstLine.setZValue(1)
            self.secondLine.setZValue(1)

            self.secondCorner.setVisible(True)
            self.thirdCorner.setVisible(True)
            self.firstLine.setVisible(True)
            self.secondLine.setVisible(True)

        else:
            self.secondCorner.setPos(newPos.x() + 10, newPos.y())
            self.thirdCorner.setPos(newPos.x() + 10, self.parent.toPort.scenePos().y())
            self.secondCorner.node.prevN().parent.setY(newPos.y())

            self.firstLine.setLine(
                self.secondCorner.node.prevN().parent.scenePos().x(),
                newPos.y(),
                self.secondCorner.scenePos().x(),
                newPos.y(),
            )
            self.secondLine.setLine(
                self.secondCorner.scenePos().x(),
                self.secondCorner.scenePos().y(),
                self.thirdCorner.scenePos().x(),
                self.thirdCorner.scenePos().y(),
            )
            self.setLine(
                self.thirdCorner.scenePos().x(),
                self.thirdCorner.scenePos().y(),
                self.endNode.parent.toPort.scenePos().x(),
                self.endNode.parent.toPort.scenePos().y(),
            )

            self.secondCorner.setZValue(100)
            self.thirdCorner.setZValue(100)
            self.firstLine.setZValue(1)
            self.secondLine.setZValue(1)

            self.secondCorner.setVisible(True)
            self.thirdCorner.setVisible(True)
            self.firstLine.setVisible(True)
            self.secondLine.setVisible(True)

    def renameConn(self):
        # dia = segmentDlg(self, self.scene().parent())
        self.scene().parent().showSegmentDlg(self)

    def printItemsAt(self):
        self.logger.debug("Items at startnode are %s", str(self.scene().items(self.line().p1())))
        self.logger.debug("Items at endnode are %s", str(self.scene().items(self.line().p2())))

        for s in self.parent.segments:
            self.logger.debug(
                "Segment in list is %s has startnode %s endnode %s",
                str(s),
                str(s.startNode.parent),
                str(s.endNode.parent),
            )

    def contextMenuEvent(self, event):
        menu = QMenu()
        a1 = menu.addAction("Rename...")
        a1.triggered.connect(self.renameConn)

        a2 = menu.addAction("Delete this connection")
        a2.triggered.connect(self.parent.deleteConnCom)

        a3 = menu.addAction("Invert this connection")
        a3.triggered.connect(self.parent.invertConnection)

        a4 = menu.addAction("Toggle name")
        a4.triggered.connect(self.parent.toggleLabelVisible)

        a5 = menu.addAction("Toggle mass flow")
        a5.triggered.connect(self.parent.toggleMassFlowLabelVisible)

        # b1 = menu.addAction('Set group ')
        # b1.triggered.connect(self.configGroup)
        # a4 = menu.addAction('Print end and start items')
        # a4.triggered.connect(self.printItemsAt)
        #
        # a5 = menu.addAction('Print corners')
        # a5.triggered.connect(self.parent.getCorners)
        #
        # a6 = menu.addAction('Print group')
        # a6.triggered.connect(self.printGroup)
        #
        # a7 = menu.addAction('Inspect')
        # a7.triggered.connect(self.inspect)
        menu.exec_(event.screenPos())

    def configGroup(self):
        GroupChooserConnDlg(self.parent, self.parent.parent)

    def printGroup(self):
        self.logger.debug(self.parent.groupName)

    def inspect(self):
        self.parent.highlightConn()
        self.parent.inspectConn()

    def setLabelVisible(self, isVisible: bool) -> None:
        self.label.setVisible(isVisible)

    def toggleLabelVisible(self) -> None:
        wasVisible = self.label.isVisible()
        self.setLabelVisible(not wasVisible)

    def setMassFlowLabelVisible(self, isVisible: bool) -> None:
        self.labelMass.setVisible(isVisible)

    def toggleMassFlowLabelVisible(self) -> None:
        wasVisible = self.labelMass.isVisible()
        self.setMassFlowLabelVisible(not wasVisible)

    def setHighlight(self, isHighlight: bool) -> None:
        if isHighlight:
            highlightPen = self._createHighlightPen()
            self.setPen(highlightPen)
        else:
            self.updateGrad()

    @staticmethod
    def _createHighlightPen() -> QPen:
        color = QColor(125, 242, 189)
        width = 4
        highlightPen = QPen(color, width)
        return highlightPen



# pylint: skip-file
# type: ignore

import random

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPen, QColor
from PyQt5.QtWidgets import QGraphicsItemGroup, QGraphicsLineItem, QMenu

from trnsysGUI.PortItem import PortItem


class HeatExchanger(QGraphicsItemGroup):
    COIL_HEIGHT_IN_PIXELS = 10
    COIL_WITH_RELATIVE_TO_TANK_WIDTH = 0.6
    PORT_ITEM_PROTRUSION_SIZE_IN_PIXELS = 8

    def __init__(
        self,
        sideNr,
        width,
        relativeInputHeight,
        relativeOutputHeight,
        storageTankWidth,
        storageTankHeight,
        parent,
        name="Untitled",
        **kwargs
    ):
        super().__init__(parent)
        self.parent = parent
        self.logger = parent.logger

        self._lines = []
        self.w = width

        self.sSide = sideNr
        self.id = self.parent.parent.parent().idGen.getID()
        self.trnsysId = self.parent.parent.parent().idGen.getTrnsysID()

        self.parent.heatExchangers.append(self)
        self.setZValue(100)

        self.port1 = PortItem("i", self.sSide, self.parent)
        self.port2 = PortItem("o", self.sSide, self.parent)

        self.parent.inputs.append(self.port1)
        self.parent.outputs.append(self.port2)

        self._setRelativeHeightsAndTankSize(
            storageTankWidth,
            storageTankHeight,
            relativeInputHeight,
            relativeOutputHeight,
        )

        if kwargs == {} or "tempHx" in kwargs:
            self.logger.debug("Creating new HeatExchanger")
            if kwargs == {}:
                self.displayName = name + str(self.id)
            elif "tempHx" in kwargs:
                self.displayName = name
            self._setPortsColor()
        else:
            if "loadedHx" in kwargs:
                self.logger.debug("Loading existing HeatExchanger")
                self.displayName = name

    def _setPortsColor(self):
        self.port1.setZValue(100)
        self.port2.setZValue(100)

        randomValue = int(random.uniform(20, 200))
        randomColor = QColor(randomValue, randomValue, randomValue)
        self.port1.innerCircle.setBrush(randomColor)
        self.port2.innerCircle.setBrush(randomColor)
        self.port1.visibleColor = randomColor
        self.port2.visibleColor = randomColor

        self._draw()

    def setId(self, newId):
        self.id = newId

    def setParent(self, p):
        self.parent = p

        if p.name is "StorageTank":
            self.logger.debug("p is a StorageTank")
            self.setParentItem(p)
        else:
            self.logger.debug(
                "A non-Storage-Tank block is trying to set parent of heatExchanger"
            )

    def contextMenuEvent(self, event):
        menu = QMenu()

        a1 = menu.addAction("Rename...")
        a1.triggered.connect(self.renameHx)

        a2 = menu.addAction("PLACEHOLDER TO FILL THIS EMPTY SPACE (Delete...)")

        a3 = menu.addAction("PLACEHOLDER TO FILL THIS EMPTY SPACE (Change position)")

        menu.exec_(event.screenPos())

    def mousePressEvent(self, event):
        self.highlightHx()
        self.logger.debug("pressed")

    def renameHx(self):
        # dia = hxDlg(self, self.scene().parent())
        self.scene().parent().showHxDlg(self)

    def rename(self, newName):
        self.displayName = newName

    def highlightHx(self):
        for ch in self.childItems():
            if isinstance(ch, QGraphicsLineItem):
                pen1 = QPen(QColor(125, 242, 189), 3)
                ch.setPen(pen1)

    def unhighlightHx(self):
        for ch in self.childItems():
            if isinstance(ch, QGraphicsLineItem):
                ch.setPen(QPen(Qt.black, 2))

    def setTankSize(self, storageTankWidth, storageTankHeight):
        self._clearDrawing()

        self._setRelativeHeightsAndTankSize(
            storageTankWidth,
            storageTankHeight,
            self.relativeInputHeight,
            self.relativeOutputHeight,
        )

        self._draw()

    def setRelativeHeights(self, relativeInputHeight: float, relativeOutputHeight: float) -> None:
        self._clearDrawing()

        self._setRelativeHeightsAndTankSize(
            self._storageTankWidth,
            self._storageTankHeight,
            relativeInputHeight,
            relativeOutputHeight,
        )

        self._draw()

    def _draw(self):
        sign = 1 if self.sSide == 0 else -1

        x, y = self._getPos()
        lineTop = QGraphicsLineItem(
            x - sign * self.PORT_ITEM_PROTRUSION_SIZE_IN_PIXELS,
            y + 0,
            x + sign * self.w,
            y + 0,
            self,
        )
        lineTop.setPen(QPen(Qt.black, 2))
        self._lines.append(lineTop)
        times = self.h / self.COIL_HEIGHT_IN_PIXELS
        self.logger.debug("Times is " + str(times))
        for i in range(int(times)):
            line1 = QGraphicsLineItem(
                x + sign * self.w,
                y + i * self.COIL_HEIGHT_IN_PIXELS,
                x + sign * (1 - self.COIL_WITH_RELATIVE_TO_TANK_WIDTH) * self.w,
                y + (i + 1 / 2) * self.COIL_HEIGHT_IN_PIXELS,
                self,
            )

            line2 = QGraphicsLineItem(
                x + sign * (1 - self.COIL_WITH_RELATIVE_TO_TANK_WIDTH) * self.w,
                y + (i + 1 / 2) * self.COIL_HEIGHT_IN_PIXELS,
                x + sign * self.w,
                y + (i + 1) * self.COIL_HEIGHT_IN_PIXELS,
                self,
            )

            line1.setPen(QPen(Qt.black, 2))
            line2.setPen(QPen(Qt.black, 2))

            self._lines.append(line1)
            self._lines.append(line2)
        lineBottom = QGraphicsLineItem(
            x - sign * self.PORT_ITEM_PROTRUSION_SIZE_IN_PIXELS,
            y + self.h,
            x + sign * self.w,
            y + self.h,
            self,
        )
        lineBottom.setPen(QPen(Qt.black, 2))
        self._lines.append(lineBottom)

    def _clearDrawing(self):
        for line in self._lines:
            self.parent.parent.parent().diagramScene.removeItem(line)

        self._lines = []

    def _setRelativeHeightsAndTankSize(
        self,
        storageTankWidth,
        storageTankHeight,
        relativeInputHeight,
        relativeOutputHeight,
    ):
        self._storageTankWidth = storageTankWidth
        self._storageTankHeight = storageTankHeight

        self.relativeOutputHeight = relativeOutputHeight
        self.relativeInputHeight = relativeInputHeight

        relativeHeight = relativeInputHeight - relativeOutputHeight
        absoluteHeight = relativeHeight * storageTankHeight

        self.h = absoluteHeight - absoluteHeight % self.COIL_HEIGHT_IN_PIXELS

        self._setPortPositions()

    def _setPortPositions(self):
        x, y = self._getPos()
        sign = 1 if self.sSide == 0 else -1
        xWithProtrusion = x - sign * self.PORT_ITEM_PROTRUSION_SIZE_IN_PIXELS
        self.port1.setPos(xWithProtrusion, y)
        self.port2.setPos(xWithProtrusion, y + self.h)

    def _getPos(self):
        x = 0 if self.sSide == 0 else self._storageTankWidth
        y = self._storageTankHeight - self.relativeInputHeight * self._storageTankHeight
        return x, y





# pylint: skip-file
# type: ignore

import sys

sys.path.append("C:\\Users\\parad\\OneDrive\\Desktop\\pytrnsys\\pytrnsys")
import pytrnsys.rsim.runParallelTrnsys as runTrnsys

import os


class RunMain:
    def runAction(self, logger, path):
        """
        It uses the self-lines readed by readConfig and change the lines from source to end.
        This is used to change a ddck file readed for another. A typical example is the weather data file
        Parameters
        ----------
        source : str
            string to be replaced in the config file in the self.lines field
        end : str
            str to replace the source in the config file in the self.lines field

        Returns
        -------
        bool
        """

        pathConfig = path
        configFile = "run.config"
        fullPath = os.path.join(path, configFile)
        logger.info("Starting RunParallelTrnsys with " + fullPath)
        try:
            runTrnsys.RunParallelTrnsys(pathConfig, configFile=configFile, transferredLogger=logger)
            return False, ""
        except ValueError as e:
            logger.error("EXCEPTION WHILE TRYING TO EXECUTE RunParallelTrnsys")
            errorStatement = ""
            for words in e.args:
                errorStatement += str(words)
            return True, errorStatement
        except OSError as e:
            logger.error("EXCEPTION WHILE TRYING TO EXECUTE RunParallelTrnsys")
            return True, str(e)
        except:
            logger.error("UNDEFINED EXCEPTION WHILE TRYING TO EXECUTE RunParallelTrnsys")
            return True, ""



__all__ = [
    "UpgradableJsonSchemaMixinVersion0",
    "UpgradableJsonSchemaMixin",
    "SerializationError",
]

import abc as _abc
import logging as _log
import typing as _tp
import uuid as _uuid
import json as _json

import dataclasses_jsonschema as _dcj

_S0 = _tp.TypeVar("_S0", bound="UpgradableJsonSchemaMixinVersion0")
_T = _tp.TypeVar("_T", bound="UpgradableJsonSchemaMixin")


_logger = _log.getLogger("root")


class SerializationError(ValueError):
    pass


class UpgradableJsonSchemaMixinVersion0(_dcj.JsonSchemaMixin):
    @classmethod
    def from_dict(
        cls: _tp.Type[_S0],
        data: _dcj.JsonDict,
        validate=True,
        validate_enums: bool = True,
    ) -> _S0:
        if cls._doesRequireVersion() and "__version__" not in data:
            raise SerializationError("No '__version__' field found.")

        if "__version__" in data:
            data = data.copy()
            actualVersion = data.pop("__version__")
            expectedVersion = str(cls.getVersion())

            if actualVersion != expectedVersion:
                raise SerializationError(
                    f"Version mismatch: expected {expectedVersion}, got {actualVersion}."
                )

        try:
            deserializedObject = super().from_dict(data, validate, validate_enums)
        except _dcj.ValidationError as error:
            raise SerializationError("Validation failed.") from error

        return _tp.cast(_S0, deserializedObject)

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,
    ) -> _dcj.JsonDict:
        data = super().to_dict(omit_none, validate, validate_enums)

        assert (
            "__version__" not in data
        ), "Serialized object dictionary from dataclasses-json already contained '__version__' key!"

        data["__version__"] = str(self.getVersion())
        return data

    @classmethod
    def json_schema(
        cls,
        embeddable: bool = False,
        schema_type: _dcj.SchemaType = _dcj.DEFAULT_SCHEMA_TYPE,
        validate_enums: bool = True,
        **kwargs,
    ) -> _dcj.JsonDict:
        schema = super().json_schema(embeddable, schema_type, validate_enums, **kwargs)

        schemaWithVersion = cls._getSchemaWithVersion(schema, embeddable)

        return schemaWithVersion

    @classmethod
    def _getSchemaWithVersion(cls, fullSchema, embeddable):
        schema = fullSchema[cls.__name__] if embeddable else fullSchema

        properties = schema.get("properties", {})
        properties = {**properties, "__version__": {"const": str(cls.getVersion())}}

        required = schema.get("required", [])
        if cls._doesRequireVersion():
            required = [*required, "__version__"]

        schema = {**schema, "properties": properties, "required": required}

        return {**fullSchema, cls.__name__: schema} if embeddable else schema

    @classmethod
    @_abc.abstractmethod
    def getVersion(cls) -> _uuid.UUID:
        """To be overwritten in subclass. Use
        .. code-block:: python

            import uuid
            print(str(uuid.uuid4()))

        to generate a version and return that version
        as a *constant* like so:

            .. code-block:: python

            def getVersion(cls):
                return uuid.UUID("41280f1d-67f5-4827-a306-4b2d71cba4c2")
        """
        raise NotImplementedError()

    @classmethod
    def _doesRequireVersion(cls) -> bool:
        return False


class UpgradableJsonSchemaMixin(UpgradableJsonSchemaMixinVersion0, _abc.ABC):
    @classmethod
    def from_dict(
        cls: _tp.Type[_T],
        data: _dcj.JsonDict,
        validate=True,
        validate_enums: bool = True,
    ) -> _T:
        if validate:
            try:
                cls._validate(data, validate_enums)
            except _dcj.ValidationError as error:
                raise SerializationError from error

        try:
            return super().from_dict(
                data, validate=False, validate_enums=validate_enums
            )
        except SerializationError:
            supersededClass = cls.getSupersededClass()

            supersededInstance = supersededClass.from_dict(
                data, validate=False, validate_enums=validate_enums
            )

            return cls.upgrade(supersededInstance)

    @classmethod
    def json_schema(
        cls,
        embeddable: bool = False,
        schema_type: _dcj.SchemaType = _dcj.DEFAULT_SCHEMA_TYPE,
        validate_enums: bool = True,
        **kwargs,
    ) -> _dcj.JsonDict:
        if embeddable:
            return cls._getEmbeddableSchema(kwargs, schema_type, validate_enums)

        return cls._getSchema(kwargs, schema_type, validate_enums)

    @classmethod
    def _getEmbeddableSchema(cls, kwargs, schemaType, validateEnums):
        fullSchema = super().json_schema(
            embeddable=True,
            schema_type=schemaType,
            validate_enums=validateEnums,
            **kwargs,
        )

        supersededClass = cls.getSupersededClass()
        fullSupersededSchema = supersededClass.json_schema(
            embeddable=True,
            schema_type=schemaType,
            validate_enums=validateEnums,
            **kwargs,
        )

        combinedFullSchema = cls._addSupersededSchemaToEmbeddable(
            fullSchema, fullSupersededSchema, schemaType, supersededClass
        )
        return combinedFullSchema

    @classmethod
    def _addSupersededSchemaToEmbeddable(
        cls, fullSchema, fullSupersededSchema, schemaType, supersededClass
    ):
        schema = fullSchema[cls.__name__]
        schemaReference = _dcj.schema_reference(schemaType, supersededClass.__name__)
        combinedSchema = {"anyOf": [schema, schemaReference]}
        combinedFullSchema = {
            **fullSchema,
            cls.__name__: combinedSchema,
            **fullSupersededSchema,
        }
        return combinedFullSchema

    @classmethod
    def _getSchema(cls, kwargs, schemaType, validateEnums):
        fullSchema = super().json_schema(
            embeddable=False,
            schema_type=schemaType,
            validate_enums=validateEnums,
            **kwargs,
        )

        supersededClass = cls.getSupersededClass()
        fullSupersededSchema = supersededClass.json_schema(
            embeddable=True,
            schema_type=schemaType,
            validate_enums=validateEnums,
            **kwargs,
        )

        combinedFullSchema = cls._addSupersededSchemaToTopLevel(
            fullSupersededSchema, fullSchema, supersededClass, schemaType
        )
        return combinedFullSchema

    @classmethod
    def _addSupersededSchemaToTopLevel(
        cls, fullSupersededSchema, fullSchema, supersededClass, schemaType
    ):
        definitions = fullSchema.get("definitions", {})

        cls._assertNoCollidingDefinitions(definitions, fullSupersededSchema)

        allDefinitions = definitions | fullSupersededSchema

        keys = {"type", "properties", "required", "description"}
        schema = {k: v for k, v in fullSchema.items() if k in keys}
        remainingSchema = {k: v for k, v in fullSchema.items() if k not in keys}

        schemaReference = _dcj.schema_reference(schemaType, supersededClass.__name__)
        combinedSchema = {"anyOf": [schema, schemaReference]}

        combinedFullSchema = {**remainingSchema, **combinedSchema, "definitions": allDefinitions}
        return combinedFullSchema

    @classmethod
    def _assertNoCollidingDefinitions(cls, definitions, fullSupersededSchema):
        collidingDefinitions = [
            (k, v)
            for k, v in fullSupersededSchema.items()
            if k in definitions and definitions[k] != v
        ]
        if collidingDefinitions:
            key, newDefinition = collidingDefinitions[0]
            currentDefinition = definitions[key]
            raise AssertionError(
                f"Colliding definition for {key}: old = {_json.dumps(currentDefinition)}, "
                f"new = {_json.dumps(newDefinition)}"
            )

    @classmethod
    def fromInstance(cls: _tp.Type[_T], instance: UpgradableJsonSchemaMixinVersion0) -> _T:
        supersededClass = cls.getSupersededClass()
        if type(instance) == supersededClass:  # pylint: disable=unidiomatic-typecheck
            return cls.upgrade(instance)

        if issubclass(supersededClass, UpgradableJsonSchemaMixin):
            partiallyUpgradedInstance = supersededClass.fromInstance(instance)

            return cls.upgrade(partiallyUpgradedInstance)

        raise ValueError(
            "`instance' isn't an instance of current or any previous version."
        )

    @classmethod
    def _doesRequireVersion(cls) -> bool:
        return True

    @classmethod
    @_abc.abstractmethod
    def getSupersededClass(cls) -> _tp.Type[UpgradableJsonSchemaMixinVersion0]:
        raise NotImplementedError()

    @classmethod
    @_abc.abstractmethod
    def upgrade(cls: _tp.Type[_T], superseded: UpgradableJsonSchemaMixinVersion0) -> _T:
        raise NotImplementedError()



# pylint: skip-file
# type: ignore

import re
import string
import os
import sys

from PyQt5.QtWidgets import QMessageBox

from trnsysGUI.Connection import Connection
from trnsysGUI.Pump import Pump
from trnsysGUI.TVentil import TVentil
from trnsysGUI.WTap_main import WTap_main
from trnsysGUI.StorageTank import StorageTank
from trnsysGUI.Connector import Connector
from trnsysGUI.TeePiece import TeePiece
from trnsysGUI.Collector import Collector


class Export(object):
    def __init__(self, objList, editor):
        self.logger = editor.logger

        self.trnsysObj = objList
        self.editor = editor
        self.maxChar = 20

        self.lineNumOfPar = 0
        for component in self.trnsysObj:
            if isinstance(component, Connection):
                numOfRelPorts = 1
            else:
                numOutputs = len(component.outputs)
                numInputs = len(component.inputs)

                if (numInputs == 0 and numOutputs == 1) or (numInputs == 1 and numOutputs == 0):
                    numOfRelPorts = 1
                else:
                    numOfRelPorts = min(numOutputs, numInputs)
            self.lineNumOfPar += numOfRelPorts
        self.numOfPar = 4 * self.lineNumOfPar + 1

    def exportBlackBox(self, exportTo="ddck"):
        f = "*** Black box component temperatures" + "\n"
        equationNr = 0
        problemEncountered = False

        for t in self.trnsysObj:
            status, equations = t.exportBlackBox()

            for equation in equations:
                f += equation + "\n"
            equationNr += len(equations)

        if exportTo == "mfs":
            lines = f.split("\n")
            f = ""
            for i in range(len(lines)):
                if "=" in lines[i]:
                    lines[i] = lines[i].split("=")[0] + "=1"
                f += lines[i] + "\n"

        f = "\nEQUATIONS " + str(equationNr) + "\n" + f + "\n"

        return problemEncountered, f

    def exportPumpOutlets(self):
        f = "*** Pump outlet temperatures" + "\n"
        equationNr = 0
        for t in self.trnsysObj:
            f += t.exportPumpOutlets()[0]
            equationNr += t.exportPumpOutlets()[1]

        f = "EQUATIONS " + str(equationNr) + "\n" + f + "\n"
        return f

    def exportMassFlows(self):  # What the controller should give
        f = "*** Massflowrates" + "\n"
        equationNr = 0

        for t in self.trnsysObj:
            f += t.exportMassFlows()[0]
            equationNr += t.exportMassFlows()[1]

        f = "EQUATIONS " + str(equationNr) + "\n" + f + "\n"

        return f

    def exportDivSetting(self, unit):
        """
        :param unit: the index of the previous unit number used.
        :return:
        """
        f = ""

        nUnit = unit
        constants = 0
        f2 = ""
        for t in self.trnsysObj:
            f2 += t.exportDivSetting1()[0]
            constants += t.exportDivSetting1()[1]

        if constants > 0:
            f = "CONSTANTS " + str(constants) + "\n"
            f += f2 + "\n"

        for t in self.trnsysObj:
            res = t.exportDivSetting2(nUnit)
            f += res[0]
            nUnit = res[1]

        return f

    def exportParametersFlowSolver(self, simulationUnit, simulationType, descConnLength):
        # If not all ports of an object are connected, less than 4 numbers will show up
        # TrnsysConn is a list containing the fromPort and twoPort in order as they appear in the export of connections
        f = ""
        f += "UNIT " + str(simulationUnit) + " TYPE " + str(simulationType) + "\n"
        f += "PARAMETERS " + str(self.numOfPar) + "\n"
        f += str(self.lineNumOfPar) + "\n"

        # exportConnsString: i/o i/o 0 0
        nameString = ""
        for t in self.trnsysObj:

            noHydraulicConnection = not isinstance(t, Connection) and not t.outputs and not t.inputs

            if noHydraulicConnection:
                continue
            else:
                ObjToCheck = t.exportParametersFlowSolver(descConnLength)
                f += ObjToCheck
                ObjToCheck = str(ObjToCheck).split(": ")[-1].rstrip()

                if len(ObjToCheck) > self.maxChar - 5:
                    nameString += ObjToCheck + "\n"

        if nameString != "":
            msgBox = QMessageBox()
            msgBox.setText(
                "The following variable names :\n%shas longer than %d characters!" % (nameString, self.maxChar - 5)
            )
            msgBox.exec_()

        tempS = f
        self.logger.debug("param solver text is ")
        self.logger.debug(f)
        t = self.convertToStringList(tempS)
        self.logger.debug("And now the ids come")

        f = "\n".join(t[0:3]) + "\n" + self.correctIds(t[3:]) + "\n"

        return f

    def convertToStringList(self, l):
        res = []
        temp = ""
        for i in l:
            if i == "\n":
                res.append(temp)
                temp = ""
            else:
                temp += i

        return res

    def findId(self, s):
        a = s[s.find("!") + 1 : s.find(" ", s.find("!"))]
        self.logger.debug(a)
        return a

    def correctIds(self, lineList):
        fileCopy = lineList[:]
        self.logger.debug("fds" + str(fileCopy))
        fileCopy = [" " + l for l in fileCopy]

        matchNumber = re.compile(r"\d+")

        counter = 0
        for line in lineList:
            counter += 1
            k = self.findId(line)
            if k != str(counter):

                descConnlen = 15
                res = fileCopy[:]
                for l in range(len(fileCopy)):
                    ids = matchNumber.findall(fileCopy[l])
                    for i in range(3):
                        if ids[i] == str(k):
                            ids[i] = str(counter)

                    fileCopyTempLine = fileCopy[l]
                    rest = fileCopyTempLine[fileCopyTempLine.find("!") :]
                    res[l] = ids[0] + " " + ids[1] + " " + ids[2] + " " + ids[3]
                    res[l] += " " * (descConnlen - len(res[l])) + rest

                fileCopy = res
                fileCopy = [l.replace("!" + str(k) + " ", "!" + str(counter) + " ") for l in fileCopy]

        return "\n".join(fileCopy)

    def exportInputsFlowSolver(self):
        f = ""
        f += "INPUTS " + str(self.lineNumOfPar) + "! for Type 935\n"

        counter = 0

        for t in self.trnsysObj:
            res = t.exportInputsFlowSolver1()
            f += res[0]
            counter += res[1]

            if counter > 8 or t == self.trnsysObj[-1]:
                f += "\n"
                counter = 0

        f += "\n*** Initial Inputs *" + "\n"

        counter2 = 0
        for t in self.trnsysObj:
            res = t.exportInputsFlowSolver2()
            f += res[0]
            counter2 += res[1]

            if counter2 > 8:
                f += "\n"
                counter2 = 0

        f += "\n\n"

        return f

    def exportOutputsFlowSolver(self, simulationUnit):
        f = ""

        abc = "ABC"

        prefix = "Mfr"
        equationNumber = 1
        nEqUsed = 1  # DC

        tot = ""

        for t in self.trnsysObj:
            noHydraulicConnection = not isinstance(t, Connection) and not t.outputs and not t.inputs

            if noHydraulicConnection:
                continue
            else:
                res = t.exportOutputsFlowSolver(prefix, abc, equationNumber, simulationUnit)
                tot += res[0]
                equationNumber = res[1]
                nEqUsed += res[2]

        head = (
            "EQUATIONS {0}	! Output up to three (A,B,C) mass flow rates of each component, positive = "
            "input/inlet, negative = output/outlet ".format(nEqUsed - 1)
        )

        f += head + "\n"
        f += tot + "\n"
        f += "\n"

        return f

    def exportPipeAndTeeTypesForTemp(self, startingUnit):
        # Prints the part of the export where the pipes, tp and div Units are printed

        f = ""
        unitNumber = startingUnit
        # typeNr1 = 929 # Temperature calculation from a tee-piece
        # typeNr2 = 931 # Temperature calculation from a pipe

        for t in self.trnsysObj:

            res = t.exportPipeAndTeeTypesForTemp(unitNumber)
            f += res[0]
            unitNumber = res[1]

        self.editor.printerUnitnr = unitNumber

        return f

    def exportPrintLoops(self):
        f = ""
        loopText = ""
        constsNr = 0
        constString = "CONSTANTS "
        suffix1 = "_save"
        string1 = "dtSim/rhoWat/"
        Pi = "3.14"
        for g in self.editor.groupList:

            loopText += "** Fluid Loop : " + g.displayName + "\n"

            loopNr = self.editor.groupList.index(g)

            diLp = "di_loop_" + str(loopNr)
            LLp = "L_loop_" + str(loopNr)
            ULp = "U_loop_" + str(loopNr)

            loopText += diLp + "=" + str(g.exportDi) + "\n"
            loopText += LLp + "=" + str(g.exportL) + "\n"
            loopText += ULp + "=" + str(g.exportU) + "\n"
            constsNr += 3
            loopText += "\n"

        f += constString + str(constsNr) + "\n"
        f += loopText + "\n"

        return f

    def exportPrintPipeLoops(self):
        f = ""
        loopText = ""
        equationString = "EQUATIONS "
        equationNr = 0

        for g in self.editor.groupList:
            loopText += "** Fluid Loop : " + g.displayName + "\n"

            loopNr = self.editor.groupList.index(g)

            diLp = "di_loop_" + str(loopNr)
            LLp = "L_loop_" + str(loopNr)
            ULp = "U_loop_" + str(loopNr)

            loopText += "**" + diLp + "=" + str(g.exportDi) + "\n"
            loopText += "**" + LLp + "=" + str(g.exportL) + "\n"
            loopText += "**" + ULp + "=" + str(g.exportU) + "\n"

            for c in g.itemList:
                if isinstance(c, Connection):
                    loopText += "*** " + c.displayName + "\n"
                    loopText += "di" + c.displayName + "=" + diLp + "\n"
                    loopText += "L" + c.displayName + "=" + LLp + "\n"
                    loopText += "U" + c.displayName + "=" + ULp + "\n"
                    equationNr += 3

            loopText += "\n"

        f += equationString + str(equationNr) + "\n"
        f += loopText + "\n"
        return f

    def exportPrintPipeLosses(self):
        f = ""
        lossText = ""
        rightCounter = 0

        for i in self.editor.groupList[0].itemList:
            if isinstance(i, Connection):
                if rightCounter == 0:
                    lossText += "P" + i.displayName + "_kW"
                else:
                    lossText += "+" + "P" + i.displayName + "_kW"
                rightCounter += 1

        if rightCounter == 0:
            lossText += "0"

        f += "*** Pipe losses\nEQUATIONS 1\nPipeLossTot=" + lossText + "\n\n"
        return f

    def exportMassFlowPrinter(self, unitnr, descLen):
        typenr = 25
        printingMode = 0
        relAbsStart = 0
        overwriteApp = -1
        printHeader = -1
        delimiter = 0
        printLabels = 1

        f = "ASSIGN " + self.editor.diagramName.split(".")[0] + "_Mfr.prt " + str(unitnr) + "\n\n"

        f += "UNIT " + str(unitnr) + " TYPE " + str(typenr)
        f += " " * (descLen - len(f)) + "! User defined Printer" + "\n"
        f += "PARAMETERS 10" + "\n"
        f += "dtSim"
        f += " " * (descLen - len(f)) + "! 1 Printing interval" + "\n"
        f += "START"
        f += " " * (descLen - len(f)) + "! 2 Start time" + "\n"
        f += "STOP"
        f += " " * (descLen - len(f)) + "! 3 Stop time" + "\n"
        f += str(unitnr)
        f += " " * (descLen - len(f)) + "! 4 Logical unit" + "\n"
        f += str(printingMode)
        f += " " * (descLen - len(f)) + "! 5 Units printing mode" + "\n"
        f += str(relAbsStart)
        f += " " * (descLen - len(f)) + "! 6 Relative or absolute start time" + "\n"
        f += str(overwriteApp)
        f += " " * (descLen - len(f)) + "! 7 Overwrite or Append" + "\n"
        f += str(printHeader)
        f += " " * (descLen - len(f)) + "! 8 Print header" + "\n"
        f += str(delimiter)
        f += " " * (descLen - len(f)) + "! 9 Delimiter" + "\n"
        f += str(printLabels)
        f += " " * (descLen - len(f)) + "! 10 Print labels" + "\n"
        f += "\n"

        s = ""
        breakline = 0
        for t in self.trnsysObj:
            if isinstance(t, Connection):
                breakline += 1
                if breakline % 8 == 0:
                    s += "\n"
                s += "Mfr" + t.displayName + " "
            if isinstance(t, TVentil) and t.isVisible():
                breakline += 1
                if breakline % 8 == 0:
                    s += "\n"
                s += "xFrac" + t.displayName + " "
        f += "INPUTS " + str(breakline) + "\n" + s + "\n" + "***" + "\n" + s + "\n\n"

        return f

    def exportTempPrinter(self, unitnr, descLen):

        typenr = 25
        printingMode = 0
        relAbsStart = 0
        overwriteApp = -1
        printHeader = -1
        delimiter = 0
        printLabels = 1

        f = "ASSIGN " + self.editor.diagramName.split(".")[0] + "_T.prt " + str(unitnr) + "\n\n"

        f += "UNIT " + str(unitnr) + " TYPE " + str(typenr)
        f += " " * (descLen - len(f)) + "! User defined Printer" + "\n"
        f += "PARAMETERS 10" + "\n"
        f += "dtSim"
        f += " " * (descLen - len(f)) + "! 1 Printing interval" + "\n"
        f += "START"
        f += " " * (descLen - len(f)) + "! 2 Start time" + "\n"
        f += "STOP"
        f += " " * (descLen - len(f)) + "! 3 Stop time" + "\n"
        f += str(unitnr)
        f += " " * (descLen - len(f)) + "! 4 Logical unit" + "\n"
        f += str(printingMode)
        f += " " * (descLen - len(f)) + "! 5 Units printing mode" + "\n"
        f += str(relAbsStart)
        f += " " * (descLen - len(f)) + "! 6 Relative or absolute start time" + "\n"
        f += str(overwriteApp)
        f += " " * (descLen - len(f)) + "! 7 Overwrite or Append" + "\n"
        f += str(printHeader)
        f += " " * (descLen - len(f)) + "! 8 Print header" + "\n"
        f += str(delimiter)
        f += " " * (descLen - len(f)) + "! 9 Delimiter" + "\n"
        f += str(printLabels)
        f += " " * (descLen - len(f)) + "! 10 Print labels" + "\n"
        f += "\n"

        s = ""
        breakline = 0
        for t in self.trnsysObj:
            if isinstance(t, Connection):
                breakline += 1
                if breakline % 8 == 0:
                    s += "\n"
                s += "T" + t.displayName + " "
        f += "INPUTS " + str(breakline) + "\n" + s + "\n" + "***" + "\n" + s + "\n\n"

        return f



# pylint: skip-file
# type: ignore

import json
import typing as tp

from trnsysGUI.AirSourceHP import AirSourceHP
from trnsysGUI.BlockItem import BlockItem
from trnsysGUI.Boiler import Boiler
from trnsysGUI.Collector import Collector
from trnsysGUI.Connection import Connection
from trnsysGUI.Connector import Connector
from trnsysGUI.Control import Control
from trnsysGUI.ExternalHx import ExternalHx
from trnsysGUI.GenericBlock import GenericBlock
from trnsysGUI.Graphicaltem import GraphicalItem
from trnsysGUI.GroundSourceHx import GroundSourceHx
from trnsysGUI.Group import Group
from trnsysGUI.HPDoubleDual import HPDoubleDual
from trnsysGUI.HeatPump import HeatPump
from trnsysGUI.HeatPumpTwoHx import HeatPumpTwoHx
from trnsysGUI.IceStorage import IceStorage
from trnsysGUI.IceStorageTwoHx import IceStorageTwoHx
from trnsysGUI.MasterControl import MasterControl
from trnsysGUI.PV import PV
from trnsysGUI.PitStorage import PitStorage
from trnsysGUI.Pump import Pump
from trnsysGUI.Radiator import Radiator
from trnsysGUI.StorageTank import StorageTank
from trnsysGUI.TVentil import TVentil
from trnsysGUI.TeePiece import TeePiece
from trnsysGUI.WTap import WTap
from trnsysGUI.WTap_main import WTap_main


class Decoder(json.JSONDecoder):
    """
    Decodes the diagram
    """

    def __init__(self, *args, **kwargs):

        self.editor = kwargs["editor"]
        self.logger = self.editor.logger
        kwargs.pop("editor")
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, arr: tp.Mapping[str, tp.Mapping[str, tp.Any]]) -> tp.Any:
        """
        This is the decoding function. object_hook seems to get executed for every sub dictionary in the json file.
        By looking for the specific key containing the name of dict elements, one can extract the needed dict.
        The name of the dicts is important because the order in which they are loaded matters (some objects depend on others)
        Parameters
        ----------
        arr

        Returns
        -------

        """

        resBlockList = []

        if ".__BlockDct__" in arr:

            resConnList = []

            sorted_items = sorted(arr.items(), key=lambda t: t[0])

            sorted_keys: tp.Sequence[str]
            sorted_values: tp.Sequence[tp.Mapping[str, tp.Any]]
            sorted_keys, sorted_values = zip(*sorted_items)

            formatted_sorted_keys = ", ".join(sorted_keys)
            self.logger.debug("keys are %s", formatted_sorted_keys)

            for i in sorted_values:
                if type(i) is not dict:
                    continue

                if ".__GroupDict__" in i:
                    self.logger.debug("Found the group dict")
                    self.logger.debug("Decoding group " + str(i["GroupName"]))

                    groupListNames = [g.displayName for g in self.editor.groupList]

                    if i["GroupName"] not in groupListNames:
                        g = Group(
                            i["Position"][0], i["Position"][1], i["Size"][0], i["Size"][1], self.editor.diagramScene
                        )
                        g.setName(i["GroupName"])

                # Adding the blocks to the scene could also be done inside Decoder now (before not possible because
                # no way of accessing the diagramView)

                elif ".__BlockDict__" in i:
                    self.logger.debug("Found a block ")
                    if i["BlockName"] == "TeePiece":
                        bl = TeePiece(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "TVentil":
                        bl = TVentil(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "Pump":
                        bl = Pump(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "Collector":
                        bl = Collector(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "HP":
                        bl = HeatPump(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "IceStorage":
                        bl = IceStorage(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "PitStorage":
                        bl = PitStorage(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "Radiator":
                        bl = Radiator(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "WTap":
                        bl = WTap(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "WTap_main":
                        bl = WTap_main(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "Connector":
                        bl = Connector(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "Boiler":
                        bl = Boiler(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "AirSourceHP":
                        bl = AirSourceHP(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "PV":
                        bl = PV(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "GroundSourceHx":
                        bl = GroundSourceHx(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )

                    # [--- new encoding
                    elif i["BlockName"] == "StorageTank":
                        bl = StorageTank(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "HeatPump":
                        bl = HeatPump(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "HPTwoHx":
                        bl = HeatPumpTwoHx(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "HPDoubleDual":
                        bl = HPDoubleDual(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "ExternalHx":
                        bl = ExternalHx(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "IceStorageTwoHx":
                        bl = IceStorageTwoHx(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "GenericBlock":
                        bl = GenericBlock(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "MasterControl":
                        bl = MasterControl(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    elif i["BlockName"] == "Control":
                        bl = Control(
                            i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"], loadedBlock=True
                        )
                    # --- new encoding]

                    elif i["BlockName"] == "GraphicalItem":
                        bl = GraphicalItem(self.editor.diagramView, loadedGI=True)

                    else:
                        bl = BlockItem(i["BlockName"], self.editor.diagramView, displayName=i["BlockDisplayName"])

                    bl.decode(i, resBlockList)

                elif ".__ConnectionDict__" in i:
                    fromPort = None
                    toPort = None

                    # Looking for the ports the connection is connected to
                    for connBl in resBlockList:
                        for p in connBl.inputs + connBl.outputs:
                            if p.id == i["PortFromID"]:
                                fromPort = p
                            if p.id == i["PortToID"]:
                                toPort = p

                    if fromPort is None:
                        self.logger.debug("Error: Did not found a fromPort")

                    if toPort is None:
                        self.logger.debug("Error: Did not found a toPort")

                    connectionKwargs = self.create_connection_kwargs(i)

                    c = Connection(fromPort, toPort, self.editor, **connectionKwargs)

                    c.decode(i)
                    resConnList.append(c)

                elif "__idDct__" in i:
                    resBlockList.append(i)
                elif "__nameDct__" in i:
                    resBlockList.append(i)
                else:
                    self.logger.debug("Error: Not recognized object in decoder, " + str(i))

            return resBlockList, resConnList

        return arr

    @staticmethod
    def create_connection_kwargs(item: tp.Mapping[str, tp.Any]) -> tp.Mapping[str, tp.Any]:
        connectionKwargs = dict(
            loadedConn=True,
            fromPortId=item["PortFromID"],
            toPortId=item["PortToID"],
            segmentsLoad=item["SegmentPositions"],
            cornersLoad=item["CornerPositions"],
        )

        if "FirstSegmentLabelPos" in item:
            connectionKwargs["labelPos"] = tuple(item["FirstSegmentLabelPos"])

        if "FirstSegmentMassFlowLabelPos" in item:
            connectionKwargs["labelMassPos"] = tuple(item["FirstSegmentMassFlowLabelPos"])

        return connectionKwargs
